package dk.brics.bisim;

import java.net.URL;
import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;

import javax.swing.text.BadLocationException;

import dk.brics.bisim.controller.AI;
import dk.brics.bisim.controller.AIAttacker;
import dk.brics.bisim.controller.AIDefender;
import dk.brics.bisim.controller.GameController;
import dk.brics.bisim.controller.GameControllerImpl;
import dk.brics.bisim.controller.MessagePasser;
import dk.brics.bisim.controller.MessagePasserImpl;
import dk.brics.bisim.controller.OptionsController;
import dk.brics.bisim.controller.SimulatorController;
import dk.brics.bisim.model.DotModel;
import dk.brics.bisim.model.GameModel;
import dk.brics.bisim.model.GlobalModel;
import dk.brics.bisim.model.GlobalModelImpl;
import dk.brics.bisim.model.OptionsModel;
import dk.brics.bisim.model.Protocol;
import dk.brics.bisim.model.SimulatorModel;
import dk.brics.bisim.model.UserProperties;
import dk.brics.bisim.view.GameView;
import dk.brics.bisim.view.MainView;

/**
 * The main class of the application.<p>
 * 
 * <b><u>Design</u></b><p>
 * 
 * Parts:
 * <ol>
 * <li>Model-View-Controller</li>
 * <li>RMI</li>
 * <li>GUI</li>
 * <li>Computer players</li>
 * </ol><p> 
 * 
 * <b>1. Model-View-Controller</b><p>
 * 
 * The application uses the MVC design pattern throughout choosing options, playing a game, and
 * simulation a CCS agent.
 * 
 * This means e.g. that the SimulatorController listens on user-events on the SimulatorView, and
 * reflects changes in the SimulatorModel.<p>
 * 
 * <b>2. RMI</b><p>
 * 
 * The game can be played remotely, where one player acts as a server, and one as a client.
 * 
 * The server chooses the input LTSs, his role, and the bisimulation type. An RMI registry is
 * then started on a fixed port (known to the client) on that host, with a security policy
 * which grants the following permissions:<br></br>
 * 
 * <tt>permission java.net.SocketPermission "localhost:*", "listen";</tt><br></br>
 * <tt>permission java.net.SocketPermission "*:*", "accept,connect,resolve";</tt><br></br>
 * <tt>permission java.io.FilePermission "<<ALL FILES>>", "read,execute";</tt><br></br>
 * <tt>permission java.util.PropertyPermission "*", "read";</tt><br></br>
 * 
 * Basically this is what is needed for reading arbitrary graph files, executing 'dot' on an arbitrary path,
 * and making remote connections.<br></br>
 * 
 * The RMI registry holds two remote objects:
 * 
 * <ul>
 * <li>A global model is used to hold dot representations of the input LTSs chosen by the server,
 * such that the client may fetch these and call dot locally to do layout.</li> 
 * 
 * <li>A message passer is used for passing game messages between players. These messages
 * obey a simple protocol and are interpreted locally to update the view and model correspondingly.
 * This means there is a local game model on both machines, which may seem redundant but has the benefit
 * of minimizing arguments and objects used in remote calls. In fact, only strings and primitive types 
 * are sent/received. A local game controller registers at the message passer to receive messages.
 * </li>
 * </ul>
 * 
 * The client may then join a game by specifying the name of the host running the game. 
 * Examples of host names are:<br></br>
 * 
 * <tt>camel25.daimi.au.dk</tt><br></br>
 * <tt>camel25 (if both machines are on the same LAN)</tt><br></br>
 * <tt>130.225.16.144</tt><br></br>
 * 
 * Whether playing as a client or as a server, a security manager is installed to govern RMI.
 * 
 * Once a server starts, a new RMI registry cannot be created on the same computer
 * until the server exits the application. In particular this means that a server
 * cannot later connect to the same computer as a client without exiting the application.
 * Of course, a server may still change role and connect as a client to another computer.<p>
 * 
 * <b>3. GUI</b><p>
 * 
 * In two-player remote games, a chat function is enabled.
 * To chat, type a message in the text field and hit return, or press the 'Send' button.
 * When the other user sends a message it is displayed below the text field.<p>
 * 
 * <b>4. Computer players</b><p>
 * 
 * One player games are possible, in which the opponent is a computer player acting as
 * either attacker or defender.
 * The AI player is started in a separate thread and also uses the message passer to
 * send status messages, but unlike remote games there is now only one game model and 
 * controller, namely the one associated with the human player.
 * 
 * When the human player quits a game, the AI thread is asked to stop running.
 * 
 * The AI strategies are based on configuration graphs.<p>
 * 
 * When a game starts, check the game type:
 * <ol>
 *  <li>2p remote, server:<br/> 
 *     Setup a RMI registry and bind global objects (globalModel,
 *     messagePasser).</li>
 *  <li>2p remote, client:<br/>
 *     Locate a remote registry at the specified host and retrieve
 *     the remote objects.</li>
 *  <li>Local game:<br/>
 *     The game controller still talks to the message passer, and if 
 *     it's a 1p game, a computer player thread is started, which also
 *     talks to the message passer.</li>
 *  </ol>
 * 
 * When a game ends:
 * <ol>
 *  <li>If an AI thread is running, tell it to quit.</li>
 *  <li>If the player is a server, unbind the remote objects.</li>
 *  <li>If the player is a client, nullify the registry.</li>
 * </ol>
 * @author Martin Mosegaard <a href=mailto:"martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class BisimulationGame {

	/**
	 * Remote object registry: Used in 2p remote games .	
	 */
	private Registry reg;
	
	private boolean hasRegistry;
	
	private UserProperties userProperties;
	private DotModel dotModel;
	private GameModel gameModel;
	private OptionsModel optionsModel;

	private GlobalModel globalModel;

	private MainView view;
	private GameView gameView;
	
	private MessagePasser messagePasser;
	private GameController gameController;
	
	/**
	 * Thread for a computer player. 
	 */
	private AI ai;
	
	private int role, clientServerRole, bisimulation;
	
	/**
	 * Starts a bisimulation game with an options model and a view.
	 * @param args Arguments are ignored
	 */
	public static void main(String[] args) {
		try {
			new BisimulationGame();
		} catch (BadLocationException e) {
			System.err.println("Unexpected problem occured while updating the trace:\n" + e.getMessage());
		}
	}
	
	/**
	 * Starts a bisimulation game, and shows the options view.
	 * @throws BadLocationException Should never be thrown
	 */
	public BisimulationGame() throws BadLocationException {
		/** Set security policy **/
		URL policyURL =	getClass().getClassLoader().getResource(JarResources.POLICY_FILE);
		System.clearProperty("java.security.policy");
		System.setProperty("java.security.policy", policyURL.toString());
		
		optionsModel = new OptionsModel();
		view = new MainView();
		hasRegistry = false;

		/**
		 * The user properties contain the dot path.
		 */
		userProperties = new UserProperties();		

		/**
		 * The dot model can create dot from CWB output
		 * and is used by the global model and the
		 * CCS simulator model.
		 */
		dotModel = new DotModel();
		
		/*** Pass 'this' to the options controller so we can start 
		 *   the game when options have been selected. ***/
		new OptionsController(this, 
							  optionsModel,
							  view.getOptionsView());
		
		/**
		 * A controller for the CCS Simulator tabbed pane.
		 */
		new SimulatorController(view.getSimulatorView(),
								new SimulatorModel(dotModel),
								userProperties.getDotPath());
	}

	/**
	 * Invoked by the options controller.
	 * Starts a new game.
	 */
	public void startGame() {
		try {
			if (optionsModel.getPlayers() == OptionsModel.TWO_PLAYER_REMOTE) {
				/*** Remote game ***/
				if (optionsModel.getClientServer() == OptionsModel.SERVER) {
					/*** This is a server ***/
					initLocal();
					initServer();
					gameView = view.getGameView(GameView.GAME_VIEW_SERVER);
				}
				else {
					/*** This is a client ***/
					try {
						initClient();
					} catch (NotBoundException e) {
						throw new RemoteException(Protocol.EX_MESSAGE_UNBOUND);
					}
					gameView = view.getGameView(GameView.GAME_VIEW_CLIENT);
				}
			}
			else {
				/*** Local game ***/
				initLocal();
				gameView = view.getGameView(GameView.GAME_VIEW_LOCAL);
			}
		
			if (!makeLTSs()) {
				System.err.println("Could not create LTSs from chosen files");
				return;
			}
			
			/*** Create state space and configuration graph in game model ***/
			gameModel.buildStateSpace();
			
			/*** Create a game controller ***/
			gameController = new GameControllerImpl(this,
													gameView,
													gameModel,
													role,
													clientServerRole,
													messagePasser);
			
			/*** Register the game controller at the message passer ***/
			messagePasser.registerGameController(gameController, clientServerRole);
			
			/*** Possibly create a computer player ***/
			if (optionsModel.getPlayers() == OptionsModel.ONE_PLAYER) {
				/*** set AI role ***/
				if (role == OptionsModel.ATTACKER) 
					ai = new AIDefender(gameModel, messagePasser);
				else 
					ai = new AIAttacker(gameModel, messagePasser);
				new Thread(ai).start();
			}
			
			/*** Last thing to do: Switch to game view ***/
			view.enableGameView();

		} catch (RemoteException e) {
			System.err.println("A remote problem occured:");
			String mes = e.getMessage();
			hasRegistry = false;
			reg = null;
			int cs = optionsModel.getClientServer();
			if (cs == OptionsModel.CLIENT) {
				if (mes.equals(Protocol.EX_MESSAGE_UNBOUND)) {
					/*** When a server quits a game, remote objects are unbound ***/
					System.err.println("A server is running at: " + optionsModel.getHost());
					System.err.println("but remote objects are not bound.");
					System.err.println("Wait for the server to start a game");
				}
				else {
					System.err.println("Check the host name and that a server is running on that host");
					System.err.println(mes);					
				}
			}
			else if (cs == OptionsModel.SERVER) {
				System.err.println("The system uses a fixed port: " + Protocol.PORT);
				System.err.println("Make sure no other server is running on this machine");
				System.err.println("since it will then occupy the port");
				System.err.println(mes);
			}
		} catch (BadLocationException e) {
			System.err.println("Unexpected problem occured while updating the trace:\n" + e.getMessage());
		} 
	
	}
	
	/**
	 * Invoked by the game controller when the (local) player wants to quit a game.
	 * Switch from game view to options view.
	 */
	public void quitGame() {
		/*** Tell computer player thread to die ***/
		if (ai != null)
			ai.kill();
			
		if (hasRegistry && optionsModel.getClientServer() == OptionsModel.SERVER) {
			try {
				reg.unbind(Protocol.GLOBAL_MODEL_NAME);
				reg.unbind(Protocol.MESSAGE_PASSER_NAME);
			} catch (AccessException e) {
				e.printStackTrace();
			} catch (RemoteException e) {
				e.printStackTrace();
			} catch (NotBoundException e) {
				e.printStackTrace();
			}
		}
		
		if (optionsModel.getClientServer() == OptionsModel.CLIENT) {
			reg = null;
		}
				
		view.enableOptionsView();
	}
	
	private void initLocal() throws RemoteException {	
		role = optionsModel.getRole();
		clientServerRole = optionsModel.getClientServer();
		bisimulation = optionsModel.getBisimulationType();

		globalModel = new GlobalModelImpl(optionsModel.getLeftLTSPath(),
					                      optionsModel.getRightLTSPath(),
					                      optionsModel.getLeftLTSFile(),
					                      optionsModel.getRightLTSFile(),
					                      role,
					                      bisimulation,
					                      dotModel);
		messagePasser = new MessagePasserImpl();		
	}
	
	private void initServer() throws RemoteException {
		/*** Install security manager ***/
		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());
			System.out.println("Security manager installed.");
		}

		/*** Create and start our own rmiregistry ***/
		if (hasRegistry) {
			reg = LocateRegistry.getRegistry(Protocol.PORT);			
		}
		else {
			reg = LocateRegistry.createRegistry(Protocol.PORT);
			hasRegistry = true;
		}
		/*** Names are unbound when a server quits ***/
		reg.rebind(Protocol.GLOBAL_MODEL_NAME,   globalModel);
		reg.rebind(Protocol.MESSAGE_PASSER_NAME, messagePasser);
		System.out.println("Remote objects registered."); 
	}
	
	private void initClient() throws RemoteException, NotBoundException {
		/*** Install security manager ***/
		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());
			System.out.println("Security manager installed.");
		}

		int serverRole = 0;

		/*** Get the registry that the server has created ***/
		String hostname = optionsModel.getHost();			
		reg = LocateRegistry.getRegistry(hostname, Protocol.PORT);
		
		/*** Try to lookup remote objects ***/
		globalModel   = (GlobalModel)   reg.lookup(Protocol.GLOBAL_MODEL_NAME);
		messagePasser = (MessagePasser) reg.lookup(Protocol.MESSAGE_PASSER_NAME);
		System.out.println("Located remote objects.");
		serverRole = globalModel.getRole();
		bisimulation = globalModel.getBisimulation();

		/*** Set the clients role as the opposite of the servers choice ***/
		if (serverRole == OptionsModel.ATTACKER)
			role = OptionsModel.DEFENDER;
		else if (serverRole == OptionsModel.DEFENDER)
			role = OptionsModel.ATTACKER;
		
		clientServerRole = OptionsModel.CLIENT;
	}
	
	/**
	 * Construct an LTSModel for the left and right LTSs
	 * and pass this to a GameController.
     *
	 * @throws RemoteException 	 */
	private boolean makeLTSs() throws RemoteException {
		/*** Try to create Grappa graphs ***/
		String leftFile  = globalModel.getLeftFile();
		String rightFile = globalModel.getRightFile();
		String leftInitialNodeName  = leftFile.substring(0, leftFile.indexOf('.'));
		String rightInitialNodeName = rightFile.substring(0, rightFile.indexOf('.'));			

		/*** Create (local) game model ***/
		gameModel = new GameModel(leftInitialNodeName, rightInitialNodeName);

		gameView.displayLTSTitles(leftFile, rightFile);		

		/*** Get dot representations and make Grappa graphs from them ***/
		return (gameModel.getLeftLTSModel().makeGraphFromDot
				 (globalModel.getLeftGraphAsDot(), userProperties.getDotPath(), bisimulation) &&
			    gameModel.getRightLTSModel().makeGraphFromDot
			     (globalModel.getRightGraphAsDot(), userProperties.getDotPath(), bisimulation));
	}
	
	
}
