package dk.brics.bisim;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;

/**
 * A class for reading resources from a jar file.
 * Inspired by <a href="http://www.javaworld.com/javaworld/javatips/jw-javatip49.html">JavaWorld Tip 49</a>
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class JarResources {

	/**
	 * Name of the jar file.
	 */
	public static final String JAR_FILE     = "bisim.jar";

	/**
	 * Name of the security policy file.
	 */
	protected static final String POLICY_FILE  = "dk/brics/bisim/bisim.policy";
	
	/**
	 * Name of the image in the options menu.
	 */
	public static final String SPLASH_FILE  = "dk/brics/bisim/view/amazing.gif";
	
	/**
	 * Name of the winner smiley image.
	 */
	public static final String WINNER_FILE  = "dk/brics/bisim/view/winner.gif";
	
	/**
	 * Name of the loser smiley image.
	 */
	public static final String LOSER_FILE   = "dk/brics/bisim/view/loser.gif";
	
	/**
	 * Name of the it-vest logo image.
	 */
	public static final String IT_VEST_FILE = "dk/brics/bisim/view/it-vest.gif";
	
	/**
	 * Name of the image shown in the Help view.
	 */
	public static final String GAME_FILE    = "dk/brics/bisim/view/game.gif";
	
	/**
	 * Name of the HTML file displayed in the Help view. 
	 */
	public static final String HELP_URL     = "help.html";

	/**
	 * Name of the HTML file displayed in the About view. 
	 */
	public static final String ABOUT_URL    = "about.html";	

	/**
	 * Tries to open a jar file, extract a given resource and
	 * return the resource data as a byte array.
	 * @param jarFileName Name of the jar file
	 * @param resourceName Name of the resource file
	 * @return A byte array representation of a jar resource
	 */
	public static byte[] getResource(String jarFileName, String resourceName) {
		byte[] byteArray = null;
		
		try {
			HashMap<String,Integer> sizes = new HashMap<String,Integer>();
			int size = 0;
			
			/*** Store the sizes of all entries ***/
			ZipFile zf = new ZipFile(jarFileName);
			Enumeration e = zf.entries();
			while (e.hasMoreElements()) {
				ZipEntry ze = (ZipEntry) e.nextElement();
				sizes.put(ze.getName(), (int) ze.getSize());
			}
			zf.close();			
			
			/*** Now get an input stream from which to read bytes ***/
			ZipInputStream zis = new ZipInputStream
			                      (new BufferedInputStream
					               (new FileInputStream(jarFileName)));
			
			/*** Find the right zip entry ***/
			ZipEntry ze;
			while ((ze = zis.getNextEntry()) != null) {
				if (ze.getName().equals(resourceName)) {
					size = (int) ze.getSize();
					if (size == -1) {
						size = sizes.get(ze.getName());
					}
					
					/*** Read data into byte array ***/
					byteArray = new byte[size];
					int chunk;
					int readBytes = 0;
					int sizeLeft = size;
					while (sizeLeft > 0) {
						chunk = zis.read(byteArray, readBytes, sizeLeft);
						if (chunk == -1) {
							break;
						}
						readBytes += chunk;
						sizeLeft -= chunk;
					}
					
					/*** Report if we didn't get what we wanted ***/
					if (readBytes != size) {
						StringBuilder sb = new StringBuilder("Resource: ");
						sb.append(resourceName).append(" not properly loaded.\nWanted: ")
						  .append(size).append(", got: ").append(readBytes);
						System.err.println(sb.toString());
					}
					break;	       
				}
			}	                      	  
		} catch (NullPointerException e) {
			e.printStackTrace();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return byteArray;
	}
	
}   
