package dk.brics.bisim.controller;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Map.Entry;

import dk.brics.bisim.equiv.CGNode;
import dk.brics.bisim.equiv.Strategy;
import dk.brics.bisim.equiv.Successor;
import dk.brics.bisim.model.GameModel;
import dk.brics.bisim.model.Protocol;
import dk.brics.bisim.view.Message;


import att.grappa.Node;

/**
 * A computer player with the attacker role.
 * The attacker is passed a partitioned state space and a boolean indicating
 * if there exists a perfect strategy.<p>
 * 
 * If so, the attacker proceeds as follows:
 * <ol>
 *  <li>Build a configuration graph based on the state space.</li>
 *  <li>At each turn:<br/>
 *      a. Lookup a strategy based on the current nodes in both LTSs.<br/>
 *      b. Choose side, action and transition from this strategy.
 *      </li></ol><p>
 *      
 * If not, use the following strategy:<br/>
 *  <ol>
 *  <li>If there is a strategy at the current configuration, use it.
 *      This allows the attacker to win if the defender makes a
 *      not-so-smart move.</li>
 *   <li>Else, create a dynamic strategy:</li>
 *   <li>Choose a side based on the number of choices on the left and the right side:<br/>
 *      a. If there is no action enabled on either side, choose the other.<br/>
 *      b. Else if there is only one choice on one side and more on the other,
 *         choose the side with more choices.<br/>
 *      c. Else just choose a random side.</li>
 *   <li> Choose an action:<br/>
 *      Random choice, for now.</li>
 *   </ol>
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class AIAttacker implements AI {
	
	private GameModel gameModel;
	private MessagePasser messagePasser;
	private HashMap<CGNode,Strategy> nodeMap;
	private Strategy currentStrategy;
	private boolean alive;
	
	/**
	 * Constructs an attacker and builds a strategy for it.
	 * @param gameModel A game model from which to get the status and the LTSs
	 * @param messagePasser Used for sending move messages to the human player
	 * @see dk.brics.bisim.equiv.ConfigurationGraph
	 */
	public AIAttacker(GameModel gameModel, MessagePasser messagePasser) {
		this.gameModel = gameModel;
		this.messagePasser = messagePasser;
		nodeMap = gameModel.getConfigurationGraph().buildPerfectAttackerStrategy();
		currentStrategy = null;
		alive = true;
	}
	
	public void run() {
		int turn;
		while (alive) {
			turn = gameModel.getTurn();
			if (turn == GameModel.TURN_ATTACKER_1) {
				/*** Set current strategy ***/
				currentStrategy = getStrategy();
				if (currentStrategy == null) {
					/*** It may be null if there is no perfect strategy ***/
					currentStrategy = getDynamicStrategy();
				}
				attackerChoseSide();
			}
			else if (turn == GameModel.TURN_ATTACKER_2) {
				attackerStep();
			}
		}
	}

	/**
	 * Look up the side in the current strategy and
	 * send a message to the message passer.
	 */
	private void attackerChoseSide() {
		int side = currentStrategy.getSide();
		String sideString;
		if (side == GameModel.LEFT_SIDE)
			sideString = Message.LEFT_SIDE;
		else 
			sideString = Message.RIGHT_SIDE;
		
		ArrayList<String> args = new ArrayList<String>();
		args.add(Integer.toString(side));
		args.add(sideString);

		try {
			/*** Appear to be thinking very hard ***/
			Thread.sleep(AI_SLEEP);						
			
			messagePasser.sendMessage(Protocol.MESSAGE_ATTACKER_CHOSE_SIDE,
					                  Protocol.encode(args));
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Look up the current strategy and send 
	 * a message to the message passer.
	 */
	private void attackerStep() {
		int side = currentStrategy.getSide();
		String action = currentStrategy.getAction();
		String nodeName = currentStrategy.getTargetNodeName();
		String sideString;
		if (side == GameModel.LEFT_SIDE)
			sideString = Message.RIGHT_SIDE;
		else
			sideString = Message.LEFT_SIDE;					
		
		ArrayList<String> args = new ArrayList<String>();						
		args.add(action);
		args.add(nodeName);
		args.add(sideString);

		try {
			/*** Appear to be thinking very hard ***/
			Thread.sleep(AI_SLEEP);

			messagePasser.sendMessage(Protocol.MESSAGE_ATTACKER_STEP,
						              Protocol.encode(args));
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}		
	}
	
	/**
	 * Choose a side based on the number of choices on either side.
	 * Strategy for choosing side:
	 * If there is no action enabled on either side, choose the other.
	 * Else if there is only one choice on one side and more on the other,
	 * choose the side with more choices.
	 * Else just choose a random side.
	 * We know that at least one side has at least one choice, otherwise
	 * the game would be over.
	 * @return A side
	 */
	private int chooseSide() {
		int leftChoices = gameModel.getLeftLTSModel().getNumberOfStrongChoices();
		int rightChoices = gameModel.getRightLTSModel().getNumberOfStrongChoices();
		int side;
		if (leftChoices == 0) {
			side = GameModel.RIGHT_SIDE;
		}
		else if (rightChoices == 0) {
			side = GameModel.LEFT_SIDE;
		}
		else if (leftChoices == 1) {
			if (rightChoices == 1) {
				side = chooseRandomSide();
			}
			else {
				side = GameModel.RIGHT_SIDE;
			}
		}
		else if (rightChoices == 1) {
			side = GameModel.LEFT_SIDE;
		}
		else side = chooseRandomSide();
		
		return side;		
	}
	
	/**
	 * Choose a random side.
	 * @return A side
	 */
	private int chooseRandomSide() {
		int side;
		if (Math.random() < 0.5)
			side = GameModel.LEFT_SIDE;
		else 
			side = GameModel.RIGHT_SIDE;
	
		return side;
	}
	
	/**
	 * Returns the strategy at the current node pair.
	 * @return The strategy for the matching pair graph node
	 */
	private Strategy getStrategy() {
		Node left = gameModel.getLeftLTSModel().getCurrentNode();
		Node right = gameModel.getRightLTSModel().getCurrentNode();
		for (Entry<CGNode,Strategy> me : nodeMap.entrySet()) {
			CGNode pgnode = me.getKey();
			if (pgnode.getLeft().equals(left) && pgnode.getRight().equals(right))
				return me.getValue();
		}

		return null;
	}
	
	/**
	 * Create a dynamic strategy. This is done when there isn't
	 * a perfect attacker strategy. 
	 */
	private Strategy getDynamicStrategy() {
		int side = chooseSide();
		ArrayList<Successor> choices;
		if (side == GameModel.LEFT_SIDE) {
			choices = gameModel.getLeftLTSModel().getStrongChoices();
		} 
		else {
			choices = gameModel.getRightLTSModel().getStrongChoices();						
		}
		/*** We know there is at least one choice ***/
		Successor suc;
		int size = choices.size();
		if (size > 1) {
			suc = choices.get((int) (Math.random() * size));
		}
		else {
			suc = choices.get(0);
		}
		String action = suc.getEdgeLabel();
		String nodeName = suc.getToNode().getName();						

		return new Strategy(side, action, nodeName);
	}

	public void kill() {
		alive = false;
	}

}
