package dk.brics.bisim.controller;

import java.rmi.RemoteException;
import java.util.ArrayList;

import dk.brics.bisim.equiv.CGNode;
import dk.brics.bisim.equiv.StateSpace;
import dk.brics.bisim.model.GameModel;
import dk.brics.bisim.model.Protocol;

import att.grappa.Node;

/**
 * A computer player with the defender role.
 * The defender is passed a partitioned state space and if there exists a
 * perfect strategy.<p>
 * 
 * If so, the defender proceeds as follows:
 * <ol>
 * <li>Get the name of the action chosen by the attacker.</li>
 * <li>Get the block number of the current node on the attacker's side.</li>
 * <li>Find a node on the defender's side, such that the node is reachable
 *      from the defender's current node through the action. (See also the
 *      method findNodeInBlock in the StateSpace class.)</li>
 *  </ol>
 *      
 * If not, use the following strategy:
 * <ol>
 * <li>Try to find a node in the same block as the attacker's current node.</li>
 * <li>If that fails, among the candidate configurations the one with the
 * greatest distance to the losing configuration (the accept node of the
 * configuration graph).
 * We can never be sure to win, so we can only attempt not to lose.</li>
 * </ol>
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class AIDefender implements AI {

	private GameModel gameModel;
	private MessagePasser messagePasser;
	
	private boolean alive;
	
	/**
	 * Create a defender and, if there isn't a winning strategy, annotate
	 * configurations with distances to the losing configuration.
	 * @param gameModel A game model from which to get the status and the LTSs
	 * @param messagePasser Used for sending move messages to the human player
	 * @see dk.brics.bisim.equiv.ConfigurationGraph
	 */
	public AIDefender(GameModel gameModel, MessagePasser messagePasser) {
		this.gameModel = gameModel;
		this.messagePasser = messagePasser;
		
		/** If we can't win, annotate the configuration graph
		 *  nodes with distances to the accept node in order 
		 *  to make smart choices.
		 */
		if (!gameModel.hasStateSpacePartition()) {
			gameModel.getConfigurationGraph().annotateNodes();			
		}
		
		alive = true;
	}
		
	public void run() {
		while (alive) {
			int turn = gameModel.getTurn();
			if (turn == GameModel.TURN_DEFENDER) {
				String attackerAction = gameModel.getAttackerAction();
				int attackerSide = gameModel.getAttackerSide();
				StateSpace stateSpace = gameModel.getStateSpace();
				Node toNode = getPerfectNode(attackerAction, attackerSide, stateSpace);
				if (toNode == null) {
					/** We did not find a node in the same block.
					 *  This is the case if there isn't a perfect strategy. **/
					toNode = getAnyNode(attackerAction, attackerSide, stateSpace);
					if (toNode != null) {
						defenderStep(toNode);
					}						
				}
				else {
					/** We found a node in the right block. 
					 *  This may also happen for non-bisimilar systems, like 
					 *  the coke machines, if the attacker wasn't smart. **/
					defenderStep(toNode);						
				}
			}
		}
	}

	/**
	 * Send a message to the message passer with
	 * the name of the node we have chosen to step to.
	 * @param toNode The node we have chosen to step to.
	 */
	private void defenderStep(Node toNode) {
		ArrayList<String> args = new ArrayList<String>();
		args.add(toNode.getName());
		try {
			/*** Appear to be thinking very hard ***/
			Thread.sleep(AI_SLEEP);

			messagePasser.sendMessage(Protocol.MESSAGE_DEFENDER_STEP,
						              Protocol.encode(args));
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Find a perfect node in the state space, i.e. a node
	 * which resides in the same block as the current node
	 * on the attacker's side.
	 * @param attackerAction The attacker's action
	 * @param attackerSide The attacker's side
	 * @return A perfect node if it exists, or null otherwise
	 */
	private Node getPerfectNode(String attackerAction, int attackerSide, StateSpace stateSpace) {
		Node toNode;
		Node fromNode;
		int blockNumber;
		if (attackerSide == GameModel.LEFT_SIDE) { 
			fromNode = gameModel.getRightLTSModel().getCurrentNode(); 
			blockNumber = stateSpace.getBlockNumber(gameModel.getLeftLTSModel().getCurrentNode()); 
			toNode = stateSpace.getNodeInBlock(GameModel.RIGHT_SIDE, fromNode, attackerAction, blockNumber);
		}
		else {
			fromNode = gameModel.getLeftLTSModel().getCurrentNode();
			blockNumber = stateSpace.getBlockNumber(gameModel.getRightLTSModel().getCurrentNode());
			toNode = stateSpace.getNodeInBlock(GameModel.LEFT_SIDE, fromNode, attackerAction, blockNumber);
		}
		return toNode;
	}

	/**
	 * Get any node which is the head of an edge from our 
     * current node, such that the edge label equals the
     * attacker's action.
	 * Get the number of choices. If there are more than one,
	 * choose the one with the greatest distance to the
	 * accept node.
	 * @param attackerAction The attacker's action
	 * @param attackerSide The attacker's side
	 * @return Any node which is the head of an edge from our 
	 *         current node, such that the edge label equals
	 *         the attacker's action.
	 */
	private Node getAnyNode(String attackerAction, int attackerSide, StateSpace stateSpace) {
		Node fromNode, attackersNode;
		ArrayList<Node> choices;
		if (attackerSide == GameModel.LEFT_SIDE) { 
			fromNode = gameModel.getRightLTSModel().getCurrentNode();
			attackersNode = gameModel.getLeftLTSModel().getCurrentNode();
			choices = stateSpace.getChoices(GameModel.RIGHT_SIDE,
											fromNode,
											attackerAction);
		}
		else {
			fromNode = gameModel.getLeftLTSModel().getCurrentNode();
			attackersNode = gameModel.getRightLTSModel().getCurrentNode();
			choices = stateSpace.getChoices(GameModel.LEFT_SIDE,
											fromNode, 
											attackerAction);
		}

		int size = choices.size();
		if (size > 1) {
			/** Try to make a move to a configuration
			 *  with greatest distance to the accept node.
			 */
			CGNode maxSoFar;
			if (attackerSide == GameModel.LEFT_SIDE) {
				Node firstNode = choices.get(0);
				maxSoFar = gameModel.getConfigurationGraph().getNode(attackersNode, firstNode);
				if (maxSoFar == null) {
					return firstNode; // No CG node means we may win
				}
				for (Node toNode : choices) {
					CGNode candidate = gameModel.getConfigurationGraph().getNode(attackersNode, toNode);
					if (candidate == null) {
						return toNode; // No CG node means we may win
					}
					else {
						int cdist = candidate.getDistance();
						int maxdist = maxSoFar.getDistance();
						if (cdist == -1) {
							return toNode;
						}
						else if (cdist > maxdist) {
							maxSoFar = candidate;
						}
					}
				}
				return maxSoFar.getRight();				
			}
			else {
				Node firstNode = choices.get(0);
				maxSoFar = gameModel.getConfigurationGraph().getNode(firstNode, attackersNode);
				if (maxSoFar == null) {
					return firstNode; // No CG node means we may win
				}
				for (Node toNode : choices) {
					CGNode candidate = gameModel.getConfigurationGraph().getNode(toNode, attackersNode);
					if (candidate == null) {
						return toNode; // No CG node means we may win
					}
					else {
						int cdist = candidate.getDistance();
						int maxdist = maxSoFar.getDistance();
						if (cdist == -1) {
							return toNode;
						}
						else if (cdist > maxdist) {
							maxSoFar = candidate;
						}
					}
				}
				return maxSoFar.getLeft();
			}
		}
		else if (size == 1) {
			return choices.get(0);
		}
		/** No choices. This should never happen, since we
		 *  should then have detected that the game is over. */
		else return null;
	}

	public void kill() {
		alive = false;
	}

}
