package dk.brics.bisim.controller;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;

import javax.swing.AbstractButton;
import javax.swing.JButton;
import javax.swing.JTextField;
import javax.swing.text.BadLocationException;

import att.grappa.Edge;
import att.grappa.Element;
import att.grappa.Graph;
import att.grappa.GrappaAdapter;
import att.grappa.GrappaBox;
import att.grappa.GrappaPanel;
import att.grappa.GrappaPoint;
import att.grappa.Node;
import att.grappa.Subgraph;

import dk.brics.bisim.BisimulationGame;
import dk.brics.bisim.model.GameModel;
import dk.brics.bisim.model.LTSModel;
import dk.brics.bisim.model.OptionsModel;
import dk.brics.bisim.model.Protocol;
import dk.brics.bisim.view.GameView;
import dk.brics.bisim.view.GameViewUpperPanel;
import dk.brics.bisim.view.Message;

/**
 * Implementation of the GameController.
 * @see GameController 
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class GameControllerImpl extends UnicastRemoteObject
                                implements GameController {
	
	private static final long serialVersionUID = 1L;

	private int myRole, clientServerRole;
	private GameModel gameModel;
	
	private GameView gameView;

	private BisimulationGame controller;
	private MessagePasser messagePasser;
	
	private boolean lonely;
	
	private GameGrappaAdapter  gameGrappaAdapter;
	private GameActionListener gameActionListener;
	private GameMouseListener  gameMouseListener;
	
	/**
	 * Constructs a game controller.
	 * @param controller A game instance
	 * @param gameView A view on which to listen on events
	 * @param gameModel A data model
	 * @param myRole As specified in the game model
	 * @param clientServerRole As specified in the game model
	 * @param messagePasser A message passer to send move messages to the other player
	 * @throws RemoteException
	 * @throws BadLocationException 
	 */
	public GameControllerImpl(BisimulationGame controller,
							  GameView gameView, 
							  GameModel gameModel,
							  int myRole,
							  int clientServerRole,
							  MessagePasser messagePasser) 
	       throws RemoteException, BadLocationException {

		this.controller = controller;
		this.gameView = gameView;
		this.gameModel = gameModel;
		this.myRole = myRole;
		this.clientServerRole = clientServerRole;
		this.messagePasser = messagePasser;

		setupListeners();
		prepareGameView();
	}
	
	private void setupListeners() {
		gameGrappaAdapter  = new GameGrappaAdapter();
		gameActionListener = new GameActionListener();
		gameMouseListener  = new GameMouseListener();

		/*** Get buttons from game view ***/
		for (AbstractButton ab : gameView.getLowerPanel().getButtons())
			ab.addActionListener(gameActionListener);
		
		/*** Get text fields from game view ***/
		for (JTextField tf : gameView.getLowerPanel().getTextFields())
			tf.addActionListener(gameActionListener);
		
		/*** Add mouse listener to LTS scroll panes ***/
		gameView.getLeftLTSPanel().getScrollPane().addMouseListener(gameMouseListener);
		gameView.getRightLTSPanel().getScrollPane().addMouseListener(gameMouseListener);
	}

	/**
	 * @throws BadLocationException Should never be thrown
	 */
	private void prepareGameView() throws BadLocationException {
		Graph leftGraph = gameModel.getLeftLTSModel().getGraph();
		Graph rightGraph = gameModel.getRightLTSModel().getGraph();
		gameView.displayLTSs(leftGraph, 
				             rightGraph,
							 gameGrappaAdapter);
		gameView.resetTrace();
		
		if (clientServerRole == OptionsModel.SERVER) {
			gameView.updateTrace(Message.WAITING, true);
			lonely = true;
		}
			
		gameModel.setTurn(GameModel.NOBODYS_TURN); // Do not begin yet
	}

	/**
	 * @param side A side
	 * @param message A message
	 * @throws BadLocationException Should never be thrown
	 */
	private void attackerChoseSide(int side, String message) throws BadLocationException {
		gameModel.setAttackerSide(side);
		gameModel.setTurn(GameModel.TURN_ATTACKER_2);
		gameView.updateTrace(message, false);
		if (myRole == OptionsModel.ATTACKER) {
			gameView.updateTrace(Message.ACTION_YOU, true, GameView.STYLE_COMMAND);
		}
		else if (myRole == OptionsModel.DEFENDER) {
			gameView.updateTrace(Message.ACTION_OTHER, true, GameView.STYLE_STATUS);
		}
		else {
			gameView.updateTrace(Message.ACTION_BOTH, true, GameView.STYLE_COMMAND);
		}
		if (side == GameModel.LEFT_SIDE) {
			gameView.getLeftLTSPanel().toggleBorder();
		}
		else {
			gameView.getRightLTSPanel().toggleBorder();
		}
	}
	
	/**
	 * @param action Attacker's action
	 * @param nodeName Name of to-node
	 * @param message A message
	 * @throws BadLocationException Should never be thrown
	 */
	private void attackerStep(String action, String nodeName, String message) throws BadLocationException {
		int attackerSide = gameModel.getAttackerSide();
		
		if (attackerSide == GameModel.LEFT_SIDE) {
			gameModel.getLeftLTSModel().stepTo(nodeName, action);
		}
		else {
			gameModel.getRightLTSModel().stepTo(nodeName, action);
		}
		
		gameView.updateTrace(action, false, GameView.STYLE_ACTION);
		
		gameModel.setAttackerAction(action);
		gameModel.setGameOver();
		if (gameModel.isGameOver()) {
			if (myRole == OptionsModel.ATTACKER) {
				gameView.updateTrace(Message.YOU_WIN, GameView.STYLE_GAMEOVER, true);
			}
			else if (myRole == OptionsModel.DEFENDER) {
				gameView.updateTrace(Message.YOU_LOSE, GameView.STYLE_GAMEOVER, false);
			}
			else {
				gameView.updateTrace(Message.ATTACKER_WINS, true, GameView.STYLE_GAMEOVER);
			}			
			
			/*** Toggle attacker side border (back to gray) ***/
			if (attackerSide == GameModel.LEFT_SIDE) {
				gameView.getLeftLTSPanel().toggleBorder();
			}
			else {
				gameView.getRightLTSPanel().toggleBorder();
			}
			gameModel.setTurn(GameModel.NOBODYS_TURN);
		}
		else {
			if (myRole == OptionsModel.DEFENDER) {
				gameView.updateTrace(Message.COUNTERMOVE_YOU, true, GameView.STYLE_COMMAND);
			}
			else if (myRole == OptionsModel.ATTACKER){
				gameView.updateTrace(Message.COUNTERMOVE_OTHER, true, GameView.STYLE_STATUS);
			}
			else {
				gameView.updateTrace(Message.COUNTERMOVE_BOTH, true, GameView.STYLE_COMMAND);
			}
			/*** Toggle both borders, since we switch ***/
			gameView.getLeftLTSPanel().toggleBorder();
			gameView.getRightLTSPanel().toggleBorder();
			gameModel.setTurn(GameModel.TURN_DEFENDER);
		}
	}
	
	/**
	 * @param nodeName Name of to-node
	 * @throws BadLocationException Should never be thrown
	 */
	private void defenderStep(String nodeName) throws BadLocationException {
		int attackerSide = gameModel.getAttackerSide();

		if (attackerSide == GameModel.LEFT_SIDE) {
			gameModel.getRightLTSModel().stepTo(nodeName);
			gameModel.getLeftLTSModel().resetCurrentEdges();
			gameView.getRightLTSPanel().toggleBorder();
		}
		else if (attackerSide == GameModel.RIGHT_SIDE) {
			gameModel.getLeftLTSModel().stepTo(nodeName);
			gameModel.getRightLTSModel().resetCurrentEdges();
			gameView.getLeftLTSPanel().toggleBorder();
		}
		
		displayConfiguration();

		gameModel.setGameOver();
		if (gameModel.isGameOver()) {
			int i = gameModel.getInfiniteGameOver();
			/** Display if an infinite game was detected **/
			if (i > GameModel.FINITE_GAME) {
				gameView.updateTrace(new StringBuilder(Message.ALREADY_SEEN).append(i).toString(),
									 true,
									 GameView.STYLE_GAMEOVER);				
			}

			/** Display who won **/
			if (myRole == OptionsModel.DEFENDER) {
				gameView.updateTrace(Message.YOU_WIN, GameView.STYLE_GAMEOVER, true);
			}
			else if (myRole == OptionsModel.ATTACKER) {
				gameView.updateTrace(Message.YOU_LOSE, GameView.STYLE_GAMEOVER, false);					
			}
			else {
				gameView.updateTrace(Message.DEFENDER_WINS, true, GameView.STYLE_GAMEOVER);
			}
			gameModel.setTurn(GameModel.NOBODYS_TURN);
		}
		else {
			if (myRole == OptionsModel.ATTACKER) {
				gameView.updateTrace(Message.SIDE_YOU, true, GameView.STYLE_COMMAND);
			}
			else if (myRole == OptionsModel.DEFENDER){
				gameView.updateTrace(Message.SIDE_OTHER, true, GameView.STYLE_STATUS);
			}
			else {
				gameView.updateTrace(Message.SIDE_BOTH, true, GameView.STYLE_COMMAND);
			}
			gameModel.setTurn(GameModel.TURN_ATTACKER_1);
		}
	}
	
	/**
	 * @throws BadLocationException Should never be thrown
	 */
	private void resetGame() throws BadLocationException {
		/*** A game starts so we're only lonely if it's a local game ***/
		lonely = (clientServerRole == OptionsModel.BOTH);
		
		gameModel.reset();
		gameModel.setTurn(GameModel.TURN_ATTACKER_1);
		
		gameView.resetTrace();
		gameView.getLeftLTSPanel().resetBorder();
		gameView.getRightLTSPanel().resetBorder();
		displayConfiguration();
		if (myRole == OptionsModel.ATTACKER) {
			gameView.updateTrace(Message.SIDE_YOU, true, GameView.STYLE_COMMAND);		
		}
		else if (myRole == OptionsModel.DEFENDER) {
			gameView.updateTrace(Message.SIDE_OTHER, true, GameView.STYLE_STATUS);		
		}
		else {
			gameView.updateTrace(Message.SIDE_BOTH, true, GameView.STYLE_COMMAND);		
		}
	}

	/**
	 * @throws BadLocationException Should never be thrown
	 */
	private void displayConfiguration() throws BadLocationException {
		String mes = new StringBuilder(Message.ROUND).append(gameModel.incAndGetRoundNumber())
		               .append(". ").toString();
		gameView.updateTrace(mes, true);
		mes = new StringBuilder(Message.CONFIGURATION).append("(").toString();
		gameView.updateTrace(mes, false);
		gameView.updateTrace(gameModel.getLeftLTSModel().getCurrentNode().getName(),
							 false,
							 GameView.STYLE_STATE);
		gameView.updateTrace(", ", false);
		gameView.updateTrace(gameModel.getRightLTSModel().getCurrentNode().getName(),
							 false,
							 GameView.STYLE_STATE);
		gameView.updateTrace(")", false);
	}
	
	public void updateStatus(int messageNumber, String message) throws RemoteException {
		ArrayList<String> decoded = Protocol.decode(messageNumber, message);
	
		switch (messageNumber) {
		case Protocol.MESSAGE_ATTACKER_CHOSE_SIDE :
			try {
				attackerChoseSide(Integer.parseInt(decoded.get(0)), 
						          decoded.get(1));
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
			break;
		case Protocol.MESSAGE_ATTACKER_STEP :
			try {
				attackerStep(decoded.get(0),
						     decoded.get(1),
						     decoded.get(2));
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
			break;
		case Protocol.MESSAGE_DEFENDER_STEP :
			try {
				defenderStep(decoded.get(0));
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
			break;
		case Protocol.MESSAGE_RESET_GAME :
			try {
				resetGame();
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
			break;
		default :
			System.err.println("Message not understood: " + message);
			break;
		}
	}

	public void quit() throws RemoteException {
		if (clientServerRole == OptionsModel.CLIENT) {
			/*** The (remote) server wants to quit ***/
			lonely = true;
			gameModel.setTurn(GameModel.NOBODYS_TURN);
			try {
				gameView.updateTrace(Message.SERVER_LEFT, true);
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
		}
		else if (clientServerRole == OptionsModel.SERVER) {
			/*** The (remote) client wants to quit ***/
			lonely = true;
			gameModel.setTurn(GameModel.NOBODYS_TURN);
			try {
				gameView.updateTrace(Message.CLIENT_LEFT, true);
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
		}
	}
	
	public void setChatMessage(String message) throws RemoteException {
		gameView.getLowerPanel()
		        .getChatMessageLabel()
		        .setText(new StringBuilder("Other player says: ").append(message).toString());
	}
	
	class GameGrappaAdapter extends GrappaAdapter {
		/***** GrappaAdapter methods for handling mouse events on the graph *****/
	
		public void grappaClicked(Subgraph subg, Element elem, GrappaPoint pt, int modifiers,
			                  int clickCount, GrappaPanel panel) {

			/*** Handle mouse clicks based on the game status ***/

			/*** React if it's our turn ***/
			int turn = gameModel.getTurn();
			if (turn == GameModel.TURN_ATTACKER_1 && myRole != OptionsModel.DEFENDER) {
				int side = 0;
				String sideString = null;
					
				if (panel == gameView.getLeftLTSPanel().getGrappaPanel()) {
					side = GameModel.LEFT_SIDE;
					sideString = Message.LEFT_SIDE;
				}
				else if (panel == gameView.getRightLTSPanel().getGrappaPanel()) {
					side = GameModel.RIGHT_SIDE;
					sideString = Message.RIGHT_SIDE;
				}

				ArrayList<String> args = new ArrayList<String>();
				args.add(Integer.toString(side));
				args.add(sideString);
				
				try {
					messagePasser.sendMessage(Protocol.MESSAGE_ATTACKER_CHOSE_SIDE,
							                  Protocol.encode(args));
				} catch (RemoteException e1) {
					e1.printStackTrace();
				}
			}	
			else if (turn == GameModel.TURN_ATTACKER_2 && myRole != OptionsModel.DEFENDER) {
				if (elem != null) {
					LTSModel ltsModel;
					String action = null;
					String nodeName = null;
					String sideString;
					
					int attackerSide = gameModel.getAttackerSide();
			
					/*** Check that attacker side and clicked panel match and set up LTS ***/ 
					if (attackerSide == GameModel.LEFT_SIDE &&
						panel == gameView.getLeftLTSPanel().getGrappaPanel()) {
						ltsModel = gameModel.getLeftLTSModel();
						sideString = Message.RIGHT_SIDE; // Defender side
					}
					else if (attackerSide == GameModel.RIGHT_SIDE &&
							 panel == gameView.getRightLTSPanel().getGrappaPanel()) {
						ltsModel = gameModel.getRightLTSModel();
						sideString = Message.LEFT_SIDE; // Defender side
					}
					else return; // No match
				
					/*** Test on clicked element and set up action name ***/
					if (elem.isEdge())  {
						Edge edge = (Edge)elem;
						action = ltsModel.getEdgeAction(edge);
						nodeName = edge.getHead().getName();
					}
					else if (elem.isNode()) {
						Node node = (Node)elem;						
						action = ltsModel.getUniqueNodeAction(node);
						nodeName = node.getName();
					}
			
					/*** If we're ready, take a step in the LTS ***/
					if (action != null) {

						ArrayList<String> args = new ArrayList<String>();						
						args.add(action);
						args.add(nodeName);
						args.add(sideString);
												
						try {
							messagePasser.sendMessage(Protocol.MESSAGE_ATTACKER_STEP,
									                  Protocol.encode(args));
						} catch (RemoteException e) {
							e.printStackTrace();
						}
					}													
				}
			}
			else if (turn == GameModel.TURN_DEFENDER && myRole != OptionsModel.ATTACKER) {
				if (elem != null) {
					LTSModel ltsModel;
					boolean isReady = false;
					int attackerSide = gameModel.getAttackerSide();
		
					/*** Check that attacker side is opposite of clicked panel and set up LTS ***/
					if (attackerSide == GameModel.LEFT_SIDE &&
						panel == gameView.getRightLTSPanel().getGrappaPanel()) {
						ltsModel = gameModel.getRightLTSModel();
					}
					else if (attackerSide == GameModel.RIGHT_SIDE &&
							 panel == gameView.getLeftLTSPanel().getGrappaPanel()) {
						ltsModel = gameModel.getLeftLTSModel();					
					}
					else return; // No match
			
					String attackerAction = gameModel.getAttackerAction();
					String nodeName = null;
			
					/*** Test on clicked element and set up action name ***/
					if (elem.isEdge())  {
						Edge edge = (Edge)elem;
						isReady = ltsModel.getMatchingEdgeAction(edge, attackerAction);
						nodeName = edge.getHead().getName();
					}
					else if (elem.isNode()) {
						Node node = (Node)elem;						
						isReady = ltsModel.getMatchingNodeAction(node, attackerAction);
						nodeName = node.getName();
					}

					/*** If we're ready, take a step in the LTS ***/
					if (isReady) {
						ArrayList<String> args = new ArrayList<String>();
						args.add(nodeName);

						try {							
							messagePasser.sendMessage(Protocol.MESSAGE_DEFENDER_STEP,
									                  Protocol.encode(args));
						} catch (RemoteException e) {
							e.printStackTrace();
						}
					}
				}
			}
		}
	
		/* (non-Javadoc)
	 * Empty implementation to avoid default behaviour.
	 * @see att.grappa.GrappaListener#grappaDragged(att.grappa.Subgraph, att.grappa.GrappaPoint, int, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.GrappaBox, att.grappa.GrappaPanel)
	 */
		public void grappaDragged(Subgraph subg, GrappaPoint currentPt, int currentModifiers,
							  Element pressedElem, GrappaPoint pressedPt,
							  int pressedModifiers, GrappaBox outline, GrappaPanel panel) {
		}
	
		/* (non-Javadoc)
	 * Empty implementation to avoid default behaviour.
	 * @see att.grappa.GrappaListener#grappaPressed(att.grappa.Subgraph, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.GrappaPanel)
	 */
		public void grappaPressed(Subgraph subg, Element elem, GrappaPoint pt, int modifiers,
			                  GrappaPanel panel) {
		}
	
		/* (non-Javadoc)
	 * Empty implementation to avoid default behaviour.
	 * @see att.grappa.GrappaListener#grappaReleased(att.grappa.Subgraph, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.GrappaBox, att.grappa.GrappaPanel)
	 */
		public void	grappaReleased(Subgraph subg, Element elem, GrappaPoint pt, int modifiers,
			                   Element pressedElem, GrappaPoint pressedPt,
			                   int pressedModifiers, GrappaBox outline, GrappaPanel panel) {
		}

	}

	/**
	 * Listen on game view elements with a possible remote effect.
	 * 
	 * @author Martin Mosegaard <a href=mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
	 *
	 */
	class GameActionListener implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			Object source = e.getSource();
			GameViewUpperPanel low = gameView.getLowerPanel();
			if (source instanceof JButton) {
				JButton target = (JButton)source;
				if (target == low.resetGameButton && !lonely) {
					try {
						messagePasser.sendMessage(Protocol.MESSAGE_RESET_GAME, "");
					} catch (RemoteException e1) {
						e1.printStackTrace();
					}
				}
				else if (target == low.quitButton) {
					/*** Notify the other player if we're first to leave ***/
					if (!lonely) {
						try {
							messagePasser.quit(clientServerRole);
						} catch (RemoteException e1) {
							e1.printStackTrace();
						}
					}
					/*** Then go ***/
					controller.quitGame();
				}
				else if (target == low.sendButton) {
					sendChatMessage();
				}
			}
			else if (source instanceof JTextField) {
				JTextField target = (JTextField)source;
				if (target == low.chatTextField) {
					sendChatMessage();
				}
			}
		}

		private void sendChatMessage() {
			/*** Get the message ***/
			String mes = gameView.getLowerPanel().chatTextField.getText();
			
			/*** Send the message to the global controller ***/
			try {							
				messagePasser.talk(clientServerRole, mes);
				gameView.getLowerPanel()
				        .getChatMessageLabel()
				        .setText(new StringBuilder("Message sent: ").append(mes).toString());
				gameView.getLowerPanel().chatTextField.setText("");
			} catch (RemoteException e1) {
				e1.printStackTrace();
			}
		}
	}
	
	class GameMouseListener implements MouseListener {

		public void mouseClicked(MouseEvent e) {
			// TODO Auto-generated method stub
		}

		public void mousePressed(MouseEvent e) {
			/*** React if it's our turn ***/
			int turn = gameModel.getTurn();
			if (turn == GameModel.TURN_ATTACKER_1 && myRole != OptionsModel.DEFENDER) {
				int side = 0;
				String sideString = null;
					
				Component target = e.getComponent();
				if (target == gameView.getLeftLTSPanel().getScrollPane()) {
					side = GameModel.LEFT_SIDE;
					sideString = Message.LEFT_SIDE;
				}
				else if (target == gameView.getRightLTSPanel().getScrollPane()) {
					side = GameModel.RIGHT_SIDE;
					sideString = Message.RIGHT_SIDE;
				}

				ArrayList<String> args = new ArrayList<String>();
				args.add(Integer.toString(side));
				args.add(sideString);
				
				try {
					messagePasser.sendMessage(Protocol.MESSAGE_ATTACKER_CHOSE_SIDE,
							                  Protocol.encode(args));
				} catch (RemoteException e1) {
					e1.printStackTrace();
				}
			}
		}

		public void mouseReleased(MouseEvent e) {
			// TODO Auto-generated method stub
		}

		public void mouseEntered(MouseEvent e) {
			// TODO Auto-generated method stub
		}

		public void mouseExited(MouseEvent e) {
			// TODO Auto-generated method stub
		}
		
	}
	
}
