package dk.brics.bisim.controller;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Iterator;

import javax.swing.AbstractButton;
import javax.swing.JButton;
import javax.swing.JRadioButton;
import javax.swing.JTextField;

import dk.brics.bisim.BisimulationGame;
import dk.brics.bisim.model.OptionsModel;
import dk.brics.bisim.view.OptionsView;

/**
 * A controller for the options view.
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class OptionsController {

	private OptionsModel optionsModel;
	private OptionsView optionsView;
	
	private BisimulationGame controller;
	
	private OptionsActionListener optionsActionListener;
	
	/**
	 * Constructs an options controller.
	 * @param controller A game instance
	 * @param optionsModel A data model for game options
	 * @param optionsView An options view on which to listen on events
	 */
	public OptionsController (BisimulationGame controller, 
							  OptionsModel optionsModel,
							  OptionsView optionsView) {
		this.controller = controller;
		this.optionsModel = optionsModel;
		this.optionsView = optionsView;
		setupListeners();
	}
	
	private void setupListeners() {
		/*** Set up listeners ***/
		
		optionsActionListener = new OptionsActionListener();
		
		/*** Get buttons from options view ***/
		for (Iterator<AbstractButton> iter = optionsView.getButtons().iterator(); iter.hasNext(); ) {
			iter.next().addActionListener(optionsActionListener);
		}
		
		/*** Get text fields from options view ***/
		for (Iterator<JTextField> iter = optionsView.getTextFields().iterator(); iter.hasNext(); ) {
			iter.next().addActionListener(optionsActionListener);
		}
		
	}
	
	private void checkProperties() {
		String mes = optionsModel.hasProperties();
		
		optionsView.displayStatus(mes);
		optionsView.startButton.setEnabled(mes.equals(""));
	}	
	
	class OptionsActionListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			Object source = e.getSource();
			if (source instanceof JRadioButton) {
				JRadioButton target = (JRadioButton)source;
				if (target == optionsView.onePlayerRadioButton) {
					optionsModel.setPlayers(OptionsModel.ONE_PLAYER);
					optionsModel.setClientServer(OptionsModel.LOCAL);
					
					optionsView.setRemotePanelEnabled(false);
					optionsView.setHostPanelEnabled(false);
					optionsView.setRolePanelEnabled(true);
					
					checkProperties();
				}
				else if (target == optionsView.twoPlayerLocalRadioButton) {
					optionsModel.setPlayers(OptionsModel.TWO_PLAYER_LOCAL);
					optionsModel.setClientServer(OptionsModel.LOCAL);
					
					optionsView.setRolePanelEnabled(false);
					optionsView.setRemotePanelEnabled(false);
					optionsView.setTabEnabled(OptionsView.ltsString, true);
					optionsView.setTabEnabled(OptionsView.bisimulationString, true);

					checkProperties();
				}
				else if (target == optionsView.twoPlayerRemoteRadioButton) {
					optionsModel.setPlayers(OptionsModel.TWO_PLAYER_REMOTE);
					
					optionsView.setRemotePanelEnabled(true);

					checkProperties();
				}
				else if (target == optionsView.strongRadioButton) {
					optionsModel.setBisimulationType(OptionsModel.BISIMULATION_STRONG);

					checkProperties();
				}
				else if (target == optionsView.weakRadioButton) {
					optionsModel.setBisimulationType(OptionsModel.BISIMULATION_WEAK);
					
					checkProperties();
				}
				else if (target == optionsView.attackerRadioButton) {
					optionsModel.setRole(OptionsModel.ATTACKER);

					optionsView.setTabEnabled(OptionsView.ltsString, true);
					optionsView.setTabEnabled(OptionsView.bisimulationString, true);
					
					checkProperties();
				}
				else if (target == optionsView.defenderRadioButton) {
					optionsModel.setRole(OptionsModel.DEFENDER);
					
					optionsView.setTabEnabled(OptionsView.ltsString, true);
					optionsView.setTabEnabled(OptionsView.bisimulationString, true);

					checkProperties();
				}
				else if (target == optionsView.serverRadioButton) {
					optionsModel.setClientServer(OptionsModel.SERVER);
					
					optionsView.setHostPanelEnabled(false);
					optionsView.setRolePanelEnabled(true);
					optionsView.setTabEnabled(OptionsView.ltsString, true);
					optionsView.setTabEnabled(OptionsView.bisimulationString, true);

					checkProperties();
				}
				else if (target == optionsView.clientRadioButton) {
					optionsModel.setClientServer(OptionsModel.CLIENT);
					
					optionsView.setRolePanelEnabled(false);
					optionsView.setTabEnabled(OptionsView.ltsString, false);
					optionsView.setTabEnabled(OptionsView.bisimulationString, false);

					optionsView.setHostPanelEnabled(true);
					
					checkProperties();
				}
			}
			else if (source instanceof JButton) {
				JButton target = (JButton)source;
				if (target == optionsView.applyButton) {
					optionsModel.setHost(optionsView.hostTextField.getText());
					
					checkProperties();					
				}
				else if (target == optionsView.leftLTSBrowseButton) {
					File file = optionsView.chooseLTS();
					if (file != null) {
						String absPath = file.getAbsolutePath();
						optionsModel.setLeftLTSFile(file.getName());
						optionsModel.setLeftLTSPath(absPath);
						optionsView.setLeftLTSPath(absPath);					
						checkProperties();
					}
				}
				else if (target == optionsView.rightLTSBrowseButton) {
					File file = optionsView.chooseLTS();
					if (file != null) {
						String absPath = file.getAbsolutePath();
						optionsModel.setRightLTSFile(file.getName());
						optionsModel.setRightLTSPath(absPath);
						optionsView.setRightLTSPath(absPath);					
						checkProperties();
					}
				}
				else if (target == optionsView.startButton) {
					optionsModel.checkRole();
					controller.startGame();
				}
				else if (target == optionsView.exitButton) {
					System.exit(0);
				}
			}
			else if (source instanceof JTextField) {
				JTextField target = (JTextField)source;
				if (target == optionsView.hostTextField) {
					optionsModel.setHost(optionsView.hostTextField.getText());
					
					checkProperties();
				}
			}
		}
	}

}
