package dk.brics.bisim.controller;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.util.HashMap;

import javax.swing.JButton;
import javax.swing.JCheckBox;

import att.grappa.Edge;
import att.grappa.Element;
import att.grappa.GrappaAdapter;
import att.grappa.GrappaBox;
import att.grappa.GrappaPanel;
import att.grappa.GrappaPoint;
import att.grappa.Node;
import att.grappa.Subgraph;

import dk.brics.bisim.model.SimulatorModel;
import dk.brics.bisim.view.SimulatorView;

/**
 * A controller for the CCS Simulator view and model.
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class SimulatorController extends GrappaAdapter {

	private SimulatorView view;
	private SimulatorModel model;
	private SimulatorActionListener actionListener;
	private SimulatorMouseListener mouseListener;
	private String dotPath;
	
	public SimulatorController(SimulatorView view, SimulatorModel model, String dotPath) {
		this.view = view;
		this.model = model;
		this.dotPath = dotPath;
		setupListeners();
	}
	
	private void setupListeners() {
		actionListener = new SimulatorActionListener();
		mouseListener = new SimulatorMouseListener();
		
		for (JButton b : view.getButtons()) {
			b.addActionListener(actionListener);
		}		
	}
	
	private void showActions() {
		HashMap<String,String> outputActions = model.getOutputActions();
		HashMap<String,String> inputActions = model.getInputActions();
		boolean hasTau = model.hasTau();
		boolean hasInput = !inputActions.isEmpty();
		boolean hasOutput = !outputActions.isEmpty();
		view.showActions(inputActions, outputActions, mouseListener);
		view.tauButton.setEnabled(hasTau);
		view.inputButton.setEnabled(hasInput);
		view.outputButton.setEnabled(hasOutput);
		view.randomButton.setEnabled(hasInput || hasTau || hasOutput);
	}

	private void showActions(String performedAction) {
		showActions();
		view.updateTrace(performedAction);
	}
	
	private void openFile() {
		File file;
		if ((file = view.openFile()) != null) {
			if (model.createLTS(file.getName(), file.getAbsolutePath(), dotPath)) {
				view.showGraph (model.getLTSModel().getGraph(), this);				
				view.resetButton.setEnabled(true);
				reset();
			}
		}
	}
	
	private void reset() {
		model.reset();
		view.resetTrace();
		showActions();
	}
	
	/***** GrappaAdapter methods for handling mouse events on the graph *****/
	
	public void grappaClicked(Subgraph subg, Element elem, GrappaPoint pt, int modifiers,
			                  int clickCount, GrappaPanel panel) {
		if (elem != null) {
			String action = null;
			String nodeName = null;
			
			if (elem.isEdge())  {
				Edge edge = (Edge)elem;
				action = model.getLTSModel().getEdgeAction(edge);
				nodeName = edge.getHead().getName();
			}
			else if (elem.isNode()) {
				Node node = (Node)elem;						
				action = model.getLTSModel().getUniqueNodeAction(node);
				nodeName = node.getName();
			}
			
			if (action != null) {
				model.stepToNode(nodeName, action);
				showActions(action);
			}
		}
	}
	
	/* (non-Javadoc)
	 * Empty implementation to avoid default behaviour.
	 * @see att.grappa.GrappaListener#grappaDragged(att.grappa.Subgraph, att.grappa.GrappaPoint, int, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.GrappaBox, att.grappa.GrappaPanel)
	 */
	public void grappaDragged(Subgraph subg, GrappaPoint currentPt, int currentModifiers,
							  Element pressedElem, GrappaPoint pressedPt,
							  int pressedModifiers, GrappaBox outline, GrappaPanel panel) {
	}
	
	/* (non-Javadoc)
	 * Empty implementation to avoid default behaviour.
	 * @see att.grappa.GrappaListener#grappaPressed(att.grappa.Subgraph, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.GrappaPanel)
	 */
	public void grappaPressed(Subgraph subg, Element elem, GrappaPoint pt, int modifiers,
			                  GrappaPanel panel) {
	}
	
	/* (non-Javadoc)
	 * Empty implementation to avoid default behaviour.
	 * @see att.grappa.GrappaListener#grappaReleased(att.grappa.Subgraph, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.Element, att.grappa.GrappaPoint, int, att.grappa.GrappaBox, att.grappa.GrappaPanel)
	 */
	public void	grappaReleased(Subgraph subg, Element elem, GrappaPoint pt, int modifiers,
			                   Element pressedElem, GrappaPoint pressedPt,
			                   int pressedModifiers, GrappaBox outline, GrappaPanel panel) {
	}
	
	class SimulatorActionListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			Object source = e.getSource();
			if (source instanceof JButton) {
				JButton target = (JButton)source;
				if (target == view.inputButton) {
					showActions(model.stepRandomInput());
				}
				else if (target == view.tauButton) {
					model.stepRandomTau();
					showActions("tau");
				}
				else if (target == view.randomButton) {
					showActions(model.stepRandomAll());
				}
				else if (target == view.outputButton) {
					showActions(model.stepRandomOutput());
				}
				else if (target == view.openFileButton) {
					openFile();
				}
				else if (target == view.resetButton) {
					reset();
				}
				else if (target == view.exitButton) {
					System.exit(0);
				}
			}
		}
	}
	
	class SimulatorMouseListener implements MouseListener {

		public void mouseClicked(MouseEvent e) {
			Component source = e.getComponent();
			if (source instanceof JCheckBox) {
				JCheckBox target = (JCheckBox)source;
				String name = target.getName();
				String action = target.getText();
				model.stepTo(name, action);
				showActions(action);
			}
		}

		public void mousePressed(MouseEvent e) {
			// TODO Auto-generated method stub			
		}

		public void mouseReleased(MouseEvent e) {
			// TODO Auto-generated method stub			
		}

		public void mouseEntered(MouseEvent e) {
			Component source = e.getComponent();
			if (source instanceof JCheckBox) {
				JCheckBox target = (JCheckBox)source;
				String name = target.getName();
				String action = target.getText();
				model.emphasizeEdge(name, action);
			}
		}

		public void mouseExited(MouseEvent e) {
			Component source = e.getComponent();
			if (source instanceof JCheckBox) {
				JCheckBox target = (JCheckBox)source;
				String name = target.getName();
				String action = target.getText();
				model.deemphasizeEdge(name, action);
			}
		}
		
	}
	
}
