package dk.brics.bisim.equiv;

/**
 * An edge in a configuration graph.
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class CGEdge {

	private int tailId, headId;
	private String leftAction, rightAction;
	private String leftNodeName, rightNodeName;	
	private boolean isLeftWeak, isRightWeak;
	
	/**
	 * Invoked when an edge to the accept node is created;
	 * then neither part is weak.
	 * @param tailId ID of the tail node
	 * @param headId ID of the head node
	 * @param leftAction Left action name 
	 * @param leftNodeName Left node name
	 * @param rightAction Right action name
	 * @param rightNodeName Right node name
	 */
	public CGEdge(int tailId, 
			      int headId,
			      String leftAction,
			      String leftNodeName,
			      String rightAction,
			      String rightNodeName) {
		this.tailId = tailId;
		this.headId = headId;
		this.leftAction = leftAction; 
		this.leftNodeName = leftNodeName;
		this.isLeftWeak = false;
		this.rightAction = rightAction;
		this.rightNodeName = rightNodeName;
		this.isRightWeak = false;
	}

	/**
	 * Constructs a configuration graph edge, in which
	 * left and right sides may be weak.
	 * @param tailId ID of the tail node
	 * @param headId ID of the head node
	 * @param leftAction Left action name
	 * @param leftNodeName Left node name
	 * @param isLeftWeak If left action is weak
	 * @param rightAction Right action name
	 * @param rightNodeName Right node name
	 * @param isRightWeak If right action is weak
	 */
	public CGEdge(int tailId, 
				  int headId,
				  String leftAction,
				  String leftNodeName,
				  boolean isLeftWeak,
				  String rightAction,
				  String rightNodeName,
				  boolean isRightWeak) {
		this.tailId = tailId;
		this.headId = headId;
		this.leftAction = leftAction; 
		this.leftNodeName = leftNodeName;
		this.isLeftWeak = isLeftWeak;
		this.rightAction = rightAction;
		this.rightNodeName = rightNodeName;
		this.isRightWeak = isRightWeak;
	}

	/**
	 * Returns the ID of the tail node.
	 * @return The ID of the tail node
	 */
	public int getTailId() {
		return tailId;
	}
	
	/**
	 * Returns the ID of the head node.
	 * @return The ID of the head node
	 */
	public int getHeadId() {
		return headId;
	}
	
	/**
	 * Returns the left action name.
	 * @return The left action name
	 */
	public String getLeftAction() {
		return leftAction;
	}

	/**
	 * Returns the right action name.
	 * @return The right action name
	 */
	public String getRightAction() {
		return rightAction;
	}

	/**
	 * Returns the left node name.
	 * @return The left node name
	 */
	public String getLeftNodeName() {
		return leftNodeName;
	}

	/**
	 * Returns the right node name.
	 * @return The right node name
	 */
	public String getRightNodeName() {
		return rightNodeName;
	}

	public String toString() {
		StringBuilder sb = new StringBuilder("<");		
		
		sb.append(tailId).append(',').append(headId).append(",(")
		  .append(leftAction).append(' ')
		  .append(leftNodeName).append(' ')
		  .append((isLeftWeak?"weak":"strong")).append("),(")
		  .append(rightAction).append(' ')
		  .append(rightNodeName).append(' ')
		  .append((isRightWeak?"weak":"strong")).append(")>");
		
		return sb.toString();
	}
	
	public boolean equals(Object other) {
		if (other instanceof CGEdge) 
			return equals((CGEdge) other);
		else 
			return false;		
	}
	
	private boolean equals(CGEdge other) {
			return (tailId == other.getTailId() &&
					headId == other.getHeadId() &&
					leftAction.equals(other.getLeftAction()) &&
					rightAction.equals(other.getRightAction()));
	}

	public int hashCode() {
		return 37 * (tailId + headId + leftAction.hashCode() + rightAction.hashCode());
	}
	
	/**
	 * Returns if left action is weak.
	 * @return If left action is weak
	 */
	public boolean isLeftWeak() {
		return isLeftWeak;
	}

	/**
	 * Returns if right action is weak.
	 * @return If right action is weak
	 */
	public boolean isRightWeak() {
		return isRightWeak;
	}
}
