package dk.brics.bisim.equiv;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.ArrayList;

import dk.brics.bisim.model.GameModel;



import att.grappa.Node;

/**
 * A graph representation of the possible combinations of states from two LTSs (S,T).
 * 
 * The possible compound configurations of a bisimulation game can be expressed by
 * a configuration graph, in which nodes represents pairs of states, one from the left
 * LTS and one from the right LTS.<p>
 * 
 * <b>Uses</b><p>
 * 
 * A configuration graph is used by an attacker to compute a winning strategy, which is a
 * path from the initial configuration (S_init, T_init) to an accept node, which represents
 * a winning configuration for the attacker.
 * All edges into the accept node represents steps that can only occur in
 * one of the two LTSs, hence will make the defender lose.<p>
 * 
 * It is also used by a non-perfect defender to approximate bisimularity of states in
 * the following way:
 * Each configuration is annotated with its distance to the accept node (which represents
 * the losing configuration for the defender). At each step, if the defender cannot go
 * to a node in the same block as the attacker's node, it chooses a node in a configuration
 * with the greatest distance (preferably a node from which the accept node cannot be
 * reached).<p>
 * 
 * It is also used to detect cycles in seen configurations, which is used to 
 * detect a possibly infinite game, in which case the defender should be declared
 * the winner of the game. A perfect attacker can always avoid configuration cycles,
 * since he does not need to visit the same state twice.<p>
 * 
 * <b>Construction</b><p>
 * 
 * A configuration graph G = (C,E) consists of a set of nodes, or configurations,  
 * and a set of edges.
 * A configuration C = (L,R, id) consists of two LTS states, one from the left LTS and
 * one from the right LTS, and an ID.
 * 
 * An edge records actions and block numbers of target states. A graph is constructed from a 
 * partitioned state space and two LTSs as follows:<p>
 * 
 * An accept node is created. This node marks a winning configuration for the
 * attacker, meaning that the attacker has taken a step in one LTS that the 
 * defender cannot mimick in the other LTS.<p>
 * 
 * Given the two initial nodes of the LTSs, an initial configuration (L0,R0) is created.
 * From this initial configuration, edges are added to possible combinations of successors,
 * and the new configurations are added to a work set, until all have been processed.
 * 
 * If there exists a unique action in one LTS, an edge to the accept node is created.
 *  
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class ConfigurationGraph {

	/**
	 * The set of edges in the graph. 
	 */
	private HashSet<CGEdge> edges;

	/**
	 * The set of nodes in the graph.
	 */
	private HashSet<CGNode> nodes;
	
	/** 
	 * The accept node marks a winning configuration
	 * for the attacker, i.e. one where the attacker
	 * has taken a step that can't be mimicked by
	 * the defender.
	 */
	private CGNode accept;
	
	/**
	 * Constructs a graph from a state space.
	 * @param ss A state space
	 */
	public ConfigurationGraph(StateSpace ss) {
		accept = new CGNode(0);
		build(ss);
	}
	
	/**
	 * Builds a graph from a state space.
	 * Construction:
	 * Start with a work set containing a pair node made from the two initial nodes.
	 * If their capabilities are equal wrt. actions (not necessarily block numbers),
	 * create out-edges. Also create new pair nodes, if these have not been processed
	 * already.
	 * @param ss A state space.
	 */
	private void build(StateSpace ss) {
		int idCount = 1; // ID 0 is taken by the accept node.

		/*** The accept node is not in the node set, since it never needs to be looked up ***/
		nodes = new HashSet<CGNode>();
		edges = new HashSet<CGEdge>();
		
		/*** A work set of nodes to be processed ***/
		ArrayList<CGNode> workSet = new ArrayList<CGNode>();
		workSet.add(new CGNode(ss.getInitialNodeLeft(), 
				   			   ss.getInitialNodeRight(),
				   			   idCount++));
		
		/*** Nodes we have already seen ***/
		ArrayList<CGNode> seen = new ArrayList<CGNode>();
		
		/*** Construct the configuration graph ***/
		CGNode cur;
		while (!workSet.isEmpty()) {
			cur = workSet.remove(0);
			nodes.add(cur);
			seen.add(cur);
			
			ArrayList<Capability> capLeft = ss.getNodeCapability(cur.getLeft());
			ArrayList<Capability> capRight = ss.getNodeCapability(cur.getRight());
			
			HashSet<String> actionsLeft = new HashSet<String>();
			for (Capability ce : capLeft) {
				actionsLeft.add(ce.getEdgeLabel());
			}
			
			HashSet<String> actionsRight = new HashSet<String>();
			for (Capability ce : capRight) {
				actionsRight.add(ce.getEdgeLabel());
			}
				
			int uniqueSide = 0;
			String uniqueAction = null;
			String uniqueNodeName = null;
			/*** See if left node contains a unique (and strong) action ***/
			for (Capability lcap : capLeft) {
				String la = lcap.getEdgeLabel();
				if (!actionsRight.contains(la) && !lcap.isWeak()) {
					uniqueSide = GameModel.LEFT_SIDE;
					uniqueAction = la;
					uniqueNodeName = lcap.getToNode().getName();
					break;
				}
			}			
			if (uniqueAction == null) {
				/*** Else, see if right node contains a unique (and strong) action ***/
				for (Capability rcap : capRight) {
					String ra = rcap.getEdgeLabel();
					if (!actionsLeft.contains(ra) && !rcap.isWeak()) {
						uniqueSide = GameModel.RIGHT_SIDE;
						uniqueAction = ra;
						uniqueNodeName = rcap.getToNode().getName();
						break;
					}
				}
			}
			
			if (uniqueAction == null) {
				/*** Else, left side and right side have same actions ***/
				for (Capability lcap : capLeft) {
					for (Capability rcap : capRight) {
						/**
						 * Both may be weak, in that case don't create
						 * an edge, since the attacker cannot choose 
						 * a weak successor.
						 */
						if (lcap.isWeak() && rcap.isWeak()) break;					
						
						String leftAction = lcap.getEdgeLabel();
						String rightAction = rcap.getEdgeLabel();
						/*** Only consider pairs of equal actions ***/						
						if (leftAction.equals(rightAction)) {
							Node leftHead, rightHead;							

							leftHead = lcap.getToNode();
							rightHead = rcap.getToNode();
							
							CGNode head = getNode(leftHead, rightHead);						
							if (head == null) {
								head = new CGNode(leftHead, rightHead, idCount++);
								nodes.add(head);
							}
							
							if (!seen.contains(head)) {
								workSet.add(head);							
							}
							
							CGEdge edge = new CGEdge(cur.getId(), 
							         head.getId(), 
							         leftAction, 
							         leftHead.getName(), 
							         lcap.isWeak(),
							         rightAction, 
							         rightHead.getName(),
							         rcap.isWeak());
							edges.add(edge);
						}
					}
				}
			}
			else {
				/*** Found a unique action, add an edge to the accept node ***/
				String emp = "";
				if (uniqueSide == GameModel.LEFT_SIDE) {
					CGEdge edge = new CGEdge(cur.getId(),
				             accept.getId(), 
				             uniqueAction, 
				             uniqueNodeName, 
				             emp, 
				             emp);
					edges.add(edge);
				}
				else { 
					CGEdge edge = new CGEdge(cur.getId(), 
				             accept.getId(), 
				             emp, 
				             emp, 
				             uniqueAction, 
				             uniqueNodeName);
					edges.add(edge);
				}
			}
		}
	}
	
	/**
	 * Annotate nodes with a distance to the accept node.
	 * If a node cannot reach the accept node, its distance
	 * is -1.
	 * Used by a computer defender to make a choice of 
	 * configuration with the greatest distance.
	 */
	public void annotateNodes() {
		/** Backtrack from accept **/
		int dist = 0;
		ArrayList<CGNode> seen = new ArrayList<CGNode>();
		ArrayList<CGNode> distSet = new ArrayList<CGNode>();
		distSet.add(accept);
		
		while (!distSet.isEmpty()) {
			ArrayList<CGNode> newDistSet = new ArrayList<CGNode>();

			for (CGNode node : distSet) {
				/** Set distance of all nodes in this round **/
				node.setDistance(dist);
				
				/** Add all reachable nodes **/
				for (CGNode tail : getTailNodes(node)) {
					if (!seen.contains(tail)) {
						newDistSet.add(tail);
						seen.add(tail);
					}
				}
			}
			distSet = newDistSet;
			dist++;
		}
	}
	
	private ArrayList<CGNode> getTailNodes(CGNode node) {
		ArrayList<CGNode> tailNodes = new ArrayList<CGNode>();
		int nodeId = node.getId();
		
		/*** Avoid self-loops ***/
		for (CGEdge e : edges) {
			int tid = e.getTailId();
			if (e.getHeadId() == nodeId && tid != nodeId) {
				CGNode tail = getNode(tid);
				if (!tailNodes.contains(tail)) {
					tailNodes.add(tail);
				}
			}
		}
		
		return tailNodes;
	}
	
	/**
	 * Used for computing an attacker strategy.
	 * Returns all in-edges of a node with the properties that:
	 * <ol>
	 * <li>They are not self-loop edges.</li>
	 * <li>At most one part of the edge (left/right) may be weak.</li>
	 * <li>If there are more edges with the same tail, one must
	 *     be chosen instead of the others. Criteria:
	 *     An edge is not chosen if there is also another edge with
	 *     the same actions going to a different node than the 'node'
	 *     parameter.</li>
	 * <li>If the tail of an edge is a node with a similar edge going
	 *     to itself (a self-loop), do not consider this edge. This is
	 *     because there is a non-deterministic defender choice at the tail
	 *     node, allowing the defender to 'hide' in a weak cycle (SCC)
	 *     on its LTS. (Similar is with respect to the left and right action names
	 *     and weaknesses.) 
	 *     Characteristics of weak cycles in terms of a configuration graph:
	 *     There are weak edges to and from each pair of configurations in 
	 *     the cycle (including self-loops for all configurations).</li>
	 * </ol>
	 * @param node A node for which to get in-edges
	 * @return A set of edges
	 */
	private HashSet<CGEdge> getProperInEdges(CGNode node) {
		HashSet<CGEdge> inEdges = new HashSet<CGEdge>();
		int nodeId = node.getId();
		
		/*** IDs of tail nodes we have seen ***/ 
		HashSet<Integer> seen = new HashSet<Integer>();

		/*** IDs of tail nodes with more edges ***/ 
		HashSet<Integer> suspicious = new HashSet<Integer>();
		
		/*** i) and ii) ***/
		for (CGEdge e : edges) {
			int tid = e.getTailId();
			if (e.getHeadId() == nodeId &&
				tid != nodeId &&
				!(e.isLeftWeak() && e.isRightWeak())) {
				inEdges.add(e);
				/*** If we have already seen the tail node, then it is suspicious ***/
				if (seen.contains(tid)) {
					suspicious.add(tid);
				}
				else {
					seen.add(tid);
				}
			}
		}
		
		/*** iii) ***/
		if (!suspicious.isEmpty()) {
			/*** For all suspicious nodes, choose one edge among the candidates ***/
			for (Integer tailId : suspicious) {
				CGNode tail = getNode(tailId);
				/*** Get the set of edges from this node to the param node ***/
				HashSet<CGEdge> marked = new HashSet<CGEdge>();
				for (CGEdge e : inEdges) {
					if (e.getTailId() == tailId) {
						marked.add(e);
					}
				}
				/*** Choose (exactly) one edge from the marked set ***/
				CGEdge choice = null;
				boolean done = false;
				Iterator<CGEdge> mi = marked.iterator();
				while (!done && mi.hasNext()) {
					CGEdge e = mi.next();
					HashSet<CGEdge> outEdges = getOutEdges(tail);
					/*** Do not consider the marked edges ***/
					outEdges.removeAll(marked);
					/** If the edge is not equal to another edge from the
					 *  suspicious node, choose it. */
					for (CGEdge oe : outEdges) {
						if (!oe.getLeftAction().equals(e.getLeftAction()) ||
							!oe.getRightAction().equals(e.getRightAction())) {
							choice = e;
							done = true;
						}
					}
				}
				/*** Remove the other edges from the inEdges set ***/				
				marked.remove(choice);
				inEdges.removeAll(marked);
			}
		}
		
		/*** iv) ***/
		if (!inEdges.isEmpty()) {
			HashSet<CGEdge> marked = new HashSet<CGEdge>();
			for (CGEdge e : inEdges) {
				CGNode tail = getNode(e.getTailId());
				/*** Get the other out-edges of the tail node ***/
				HashSet<CGEdge> tailOutEdges = getOutEdges(tail);
				tailOutEdges.remove(e);
				/** See if one of them is a self-loop edge and
				 *  identical wrt. actions and weakness **/
				for (CGEdge oe : tailOutEdges) {
					if (oe.getTailId() == oe.getHeadId() &&
						oe.getLeftAction().equals(e.getLeftAction()) &&
						oe.getRightAction().equals(e.getRightAction()) &&
						oe.isLeftWeak() == e.isLeftWeak() &&
						oe.isRightWeak() == e.isRightWeak()) {
						marked.add(e);
						break;
					}
				}			
			}
			/*** Remove the marked edges from the inEdges set ***/				
			inEdges.removeAll(marked);		
		}
		
		return inEdges;
	}

	/**
	 * Returns all out-edges of a node.
	 * @param node A node for which to get out-edges
	 * @return A set of edges
	 */
	private HashSet<CGEdge> getOutEdges(CGNode node) {
		HashSet<CGEdge> outEdges = new HashSet<CGEdge>();
		int nodeId = node.getId();
		
		for (CGEdge e : edges) 
			if (e.getTailId() == nodeId)
				outEdges.add(e);
		
		return outEdges;
	}
	
	private CGNode getNode(int nodeId) {
		for (CGNode node : nodes)
			if (node.getId() == nodeId)
				return node;
		
		return null;
	}
	
	/**
	 * Test if there already exists a CGNode with these left and right nodes.
	 * @param left A left node
	 * @param right A right node
	 * @return The node, if it exists, otherwise <tt>null</tt>
	 */
	public CGNode getNode(Node left, Node right) {
		for (CGNode n : nodes)
			if (n.getLeft().equals(left) && n.getRight().equals(right))
				return n;
		
		return null;
	}
	
	/**
	 * Computes a strategy for the attacker.
	 * A strategy consists of, at each configuration (pair of nodes),
	 * a side (left/right), an action, and the name of the target state.
	 * Construction:
	 * <ul>
	 * <li>Start with the accept node of the pair graph, indicating that a
	 * winning move has been made by the attacker.</li>
	 * <li>Back-track on all in-edges, until the initial node is reached. At each step,
	 * choose an edge and a strategy as follows:</li>
	 * <li>If one part (left/right) of the edge is empty, choose the other part. This is
	 * the case for all edges into the accept node.</li>
	 * <li>Else if one part of the edge also occurs on another edge from the tail node,
	 * choose the other part.</li>
	 * <li>Otherwise just choose the left part. It shouldn't matter.</li>
	 * </ul>
	 * @return An attacker strategy as a map from configurations to strategies
	 */
	public HashMap<CGNode,Strategy> buildPerfectAttackerStrategy() {
		HashMap<CGNode,Strategy> nodeMap = new HashMap<CGNode,Strategy>();
		
		/*** The work set contains edges, the tails of which need a strategy ***/
		ArrayList<CGEdge> workSet = new ArrayList<CGEdge>();
		HashSet<CGEdge> inEdges = getProperInEdges(accept);
		workSet.addAll(inEdges);

		/*** These are the nodes we have seen (as the tail of an edge) ***/
		ArrayList<CGNode> seen = new ArrayList<CGNode>();
				
		CGEdge currentEdge;
		CGNode currentNode;
		Strategy strategy;
		String leftAction, leftNodeName, rightAction, rightNodeName;

		/** Back-track from the accept node.  
		 *  Even if there isn't a winning strategy, there may still be
		 *  some in-edges to the accept node, corresponding to a 
		 *  not-so-smart move by the defender - we want to exploit
		 *  that.
		 * ***/
		while (!workSet.isEmpty()) {
			currentEdge = workSet.remove(0);
			currentNode = getNode(currentEdge.getTailId());				
			
			/*** Only create a strategy if it has not been done yet ***/
			if (!seen.contains(currentNode)) {

				seen.add(currentNode);
			
				/*** Create a strategy: side, action, and target node name ***/
				strategy = null;
				leftAction    = currentEdge.getLeftAction();
				leftNodeName  = currentEdge.getLeftNodeName();
				rightAction   = currentEdge.getRightAction();
				rightNodeName = currentEdge.getRightNodeName();

				/*** If the left part of edge is empty, choose the right, and vice-versa ***/
				if (leftAction.equals("")) {
					strategy = new Strategy(GameModel.RIGHT_SIDE, rightAction, rightNodeName);
				}
				else if (rightAction.equals("")) {
					strategy = new Strategy(GameModel.LEFT_SIDE, leftAction, leftNodeName);
				}
				/*** If either part is weak, choose the other ***/
				else if (currentEdge.isLeftWeak()) {
					strategy = new Strategy(GameModel.RIGHT_SIDE, rightAction, rightNodeName);			
				}
				else if (currentEdge.isRightWeak()) {
					strategy = new Strategy(GameModel.LEFT_SIDE, leftAction, leftNodeName);			
				}
				else {			
					/*** If the left part of edge is also in another edge,
					 *   choose the right, and vice-versa.  ***/
					HashSet<CGEdge> curOut = getOutEdges(currentNode);
					curOut.remove(currentEdge);
					
					for (CGEdge e : curOut) {
						if (e.getLeftAction().equals(leftAction) &&
							e.getLeftNodeName().equals(leftNodeName)) {
							strategy = new Strategy(GameModel.RIGHT_SIDE, rightAction, rightNodeName);
							break;
						}
						else if (e.getRightAction().equals(rightAction) &&
								 e.getRightNodeName().equals(rightNodeName)) {
							strategy = new Strategy(GameModel.LEFT_SIDE, leftAction, leftNodeName);
							break;
						}
					}					
				
					/*** Else, just choose the left side (doesn't matter) ***/
					if (strategy == null) {
						strategy = new Strategy(GameModel.LEFT_SIDE, leftAction, leftNodeName);				
					}
				}
				
				/*** Attach the strategy to the current node ***/
				nodeMap.put(currentNode, strategy);
				
				/*** Update work-set: Get all in-edges, that are not self-loop edges ***/
				inEdges = getProperInEdges(currentNode);			
				/*** Add any edge with a tail node that does not yet have a strategy ***/
				for (CGEdge e : inEdges) {
					CGNode tail = getNode(e.getTailId());
					if (!seen.contains(tail)) {
						workSet.add(e);
					}
				}
			}
		}
		return nodeMap;
	}
	
	/**
	 * Used by a defender to get the set of nodes
	 * which have edges to the accept node.
	 * @return The set of suspicious nodes
	 */
	public HashSet<CGNode> getSuspiciousNodes() {
		HashSet<CGNode> nodes = new HashSet<CGNode>();
		int nid = accept.getId();
		
		for (CGEdge e : edges) {
			if (e.getHeadId() == nid) {
				CGNode tail = getNode(e.getTailId());
				if (tail != null)
					nodes.add(tail);
			}
		}
		
		return nodes;
	}
	
	public String toString() {
		StringBuilder sb = new StringBuilder("[ConfigurationGraph\n Edges:");

		sb.append(edges.size())
		  .append("\n Nodes:")
		  .append(nodes.size()+1) // Including the accept node
		  .append("\n]");
		
		return sb.toString();
	}
		
}
