package dk.brics.bisim.equiv;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.ArrayList;

import dk.brics.bisim.model.GameModel;



import att.grappa.Node;

/**
 * A representation of all nodes of two LTSs (S,T) partitioned into
 * one or more blocks.<p> 
 * 
 * The state space of nodes in the two LTSs can be partitioned according to the 
 * capabilities of a node. This is an approach also taken in the Edinburgh CWB.
 * If the two systems are bisimilar, each block in the partitioned state space contains
 * at least one node from both LTSs. In particular, the initial nodes are in the
 * same block.
 * If the two systems are not bisimilar, the initial nodes are not in the same
 * block.<p>
 * 
 * Partitioning a state space is done by comparing the capability of nodes: Nodes with
 * the same capability end up in the same block, and new blocks are created when
 * differences are encountered.<p>
 * 
 * The capability of a node is a set of pairs <a,n> of actions from that node and a
 * block number of the target node. Note that while partitioning the state space, the
 * capabilities of a node change due to changing block numbers.<p>
 * 
 * Initially a state space contains one block, which contains all nodes.
 * A successor map is used for partitioning the state space. This map is created in
 * the LTSModel, based on the chosen bisimulation type.
 * After partitioning the state space, two processes are bisimilar iff the initial
 * nodes of S and T lie in the same block.
 * From a partitioned state space, a computer defender or attacker may create a 
 * winning strategy.<p>
 * 
 * This class contains a method 'partition' for partitioning the state space,
 * which is inspired by the partition function in the source code
 * of the Edinburgh Concurrency Workbench.
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class StateSpace {

	/**
	 * Blocks making up the partition.
	 */
	private ArrayList<Block> blocks;
	
	/**
	 * A map from nodes to their capabilities. 
	 */
	private HashMap<Node,ArrayList<Capability>> capabilityMap;
	
	/**
	 * The initial nodes of the left LTS. 
	 */
	private Node initialNodeLeft;
	
	/**
	 * The initial node of the right LTS.
	 */
	private Node initialNodeRight;
	
	private HashMap<Node,ArrayList<Successor>> leftSuccessorMap;
	private HashMap<Node,ArrayList<Successor>> rightSuccessorMap;

	/**
	 * Constructs a state space based on a successor map, but 
	 * does not partition it.
	 * @param initialNodeLeft The initial node of the left LTS
	 * @param initialNodeRight The initial node of the right LTS
	 * @param leftSuccessorMap The successor map of the left LTS
	 * @param rightSuccessorMap The successor map of the right LTS
	 */
	public StateSpace(Node initialNodeLeft, 
					  Node initialNodeRight,
					  HashMap<Node,ArrayList<Successor>> leftSuccessorMap,
					  HashMap<Node,ArrayList<Successor>> rightSuccessorMap) {
		this.initialNodeLeft = initialNodeLeft;
		this.initialNodeRight = initialNodeRight;
		this.leftSuccessorMap = leftSuccessorMap;
		this.rightSuccessorMap = rightSuccessorMap;
		blocks = new ArrayList<Block>();
		capabilityMap = new HashMap<Node,ArrayList<Capability>>();
		
		/*** Initially one block is created, which contains all nodes ***/
		Set<Node> lnodes = leftSuccessorMap.keySet();
		Set<Node> rnodes = rightSuccessorMap.keySet();
		HashSet<Node> nodes = new HashSet<Node>(lnodes);
		nodes.addAll(rnodes);		
		createBlock(nodes);
	}
		
	/**
	 * Creates a block with running block number
	 * and a set of nodes, adds the block to the
	 * block ArrayList, and increments running
	 * block number.
	 * @param nodes Nodes in the new block
	 */
	private void createBlock(Set<Node> nodes) {	
		blocks.add(new Block(nodes, blocks.size()));
	}
	
	/**
	 * Retrieve the capability of a node.
	 * Assumes that the state space has been partitioned.
	 * @param node A node
	 * @return The nodes capability
	 */
	public ArrayList<Capability> getNodeCapability(Node node) {
		return capabilityMap.get(node);
	}
	
	/**
	 * Lookup if the capability has already been computed.
	 * If not, compute it and tabulate it for later lookup.
	 * A capability records all outgoing actions as well
	 * as the block number the target nodes.
	 * @param node A node
	 * @return The nodes capability
	 */
	private ArrayList<Capability> computeNodeCapability(Node node) {
		ArrayList<Capability> cap = capabilityMap.get(node);

		if (cap == null) {
			/*** Not found, so compute it ***/
			cap = new ArrayList<Capability>();

			/*** A node is in either the left or the right successor map ***/
			ArrayList<Successor> successors;
			if ((successors = leftSuccessorMap.get(node)) == null) {
				for (Successor suc : rightSuccessorMap.get(node)) {
					Node toNode = suc.getToNode();
					int blockNumber = getBlockNumber(toNode);
					if (blockNumber != -1) 
						cap.add(new Capability(suc.getEdgeLabel(), toNode, suc.isWeak(), blockNumber));
				}
			}
			else {
				for (Successor suc : successors) {
					Node toNode = suc.getToNode();
					int blockNumber = getBlockNumber(toNode);		
					if (blockNumber != -1) 
						cap.add(new Capability(suc.getEdgeLabel(), toNode, suc.isWeak(), blockNumber));
				}
			}
			
			/*** Record the capability for later lookup ***/
			capabilityMap.put(node, cap);
		}
		
		return cap;
	}
	
	/**
	 * Returns the block number of a node.
	 * @param node A node
	 * @return The block number of the node, or -1 if it is not found
	 */
	public int getBlockNumber(Node node) {
		for (Block b : blocks) {
			if (b.getNodes().contains(node)) 
				return b.getBlockNumber();
		}		
		return -1; // Not found
	}
	
	/**
	 * Checks if all nodes in a block have equal
	 * capabilities. This is done by comparing the 
	 * capability of one node (the first) with the
	 * capabilities of the other nodes, and marking
	 * nodes that differ. If the set of marked nodes is
	 * non-empty, it forms a new block, and the marked 
	 * nodes are removed from the current block.
	 * Capabilities are equal when,
	 * for all actions a and block numbers m;
	 * this.contains(a,m) ==> other.contains(a,m) and
	 * other.contains(a,m) ==> this.contains(a,m)
	 * @param block The block (partition) to consider
	 * @return True if we should split
	 */
	private boolean checkBlock(Block block) {
		/*** Look at one (the first) node in this block ***/
		ArrayList<Node> blockNodes = new ArrayList<Node>(block.getNodes());
		Node first = blockNodes.remove(0);
		ArrayList<Capability> cap = computeNodeCapability(first);
		
		HashSet<Node> markedNodes = new HashSet<Node>();
		
		/*** See if other nodes have equal capabilities ***/
		for (Node node : blockNodes) {
			ArrayList<Capability> cap2 = computeNodeCapability(node);
			if (!(cap.containsAll(cap2) &&
			      cap2.containsAll(cap))) {
				/*** If not, mark the other node ***/
				markedNodes.add(node);
			}
		}
		
		/*** We should split the block if there are any marked nodes ***/
		if (markedNodes.isEmpty()) {
			/*** If we should not split, return false to signal the block is ok ***/
			return false;
		}
		else {
			/*** If we should split, put the marked nodes in a new block 
			 *   and remove them from the old block.
			 *   Also, invalidate the capability map. ***/
			block.getNodes().removeAll(markedNodes);
			createBlock(markedNodes);
			capabilityMap.clear();
				
			return true;
		}
	}
	
	/**
	 * Partitions the state space according to
	 * strong bisimulation and returns
	 * true if the two initial nodes are in
	 * the same block (partition). 
	 * @return True if the initial nodes lie in the
	 *         same partition
	 */
	public boolean partition() {
		/*** First partition the state space.
		 * checkBlock returns true whenever a new block (partition) has
		 * been created, which means we must consider all blocks again,
		 * since capabilities may have changed.
		 * ***/	
		boolean unstable = true;
		while (unstable) {
			unstable = false;
			for (Block block : blocks)
				if (checkBlock(block)) {
					unstable = true;
					break; // Check all blocks again
				}
		}
		
		/*** Then return if the initial nodes are still in the same block ***/		
		return getBlockNumber(initialNodeLeft) ==
			   getBlockNumber(initialNodeRight);
	}

	public String toString() {
		StringBuilder sb = new StringBuilder("[StateSpace\n Initial nodes: (");
	
		/*** Write initial nodes ***/
		sb.append(initialNodeLeft.getName())
		  .append(") (")
		  .append(initialNodeRight.getName())
		  .append(")\n Blocks:");
		
		/*** Write blocks ***/
		for (Block b : blocks) 
			sb.append("\n\t").append(b.toString());
		
		sb.append("\n]");
		
		return sb.toString();	
	}

	private Block getBlockByNumber(int blockNumber) {
		for (Block b : blocks) {
			if (b.getBlockNumber() == blockNumber) return b;
		}
		return null;
	}

	/**
	 * Find a node that is head of an edge with label=action
	 * and which has tail=from, such that the node is in the
	 * block with number=blockNumber.
	 * @param side The side (left/right)
	 * @param from The from-node
	 * @param action The action name
	 * @param blockNumber The block number of the to-node
	 * @return A node
	 */
	public Node getNodeInBlock(int side, Node from, String action, int blockNumber) {
		Block block = getBlockByNumber(blockNumber);
		
		if (block != null) {
			/* Look in the right successor map */
			if (side == GameModel.LEFT_SIDE) {
				for (Successor suc : leftSuccessorMap.get(from)) {
					if (suc.getEdgeLabel().equals(action)) {
						Node toNode = suc.getToNode();
						if (block.getNodes().contains(toNode))
							return toNode;
					}
				}
			}
			else {
				for (Successor suc : rightSuccessorMap.get(from)) {
					if (suc.getEdgeLabel().equals(action)) {
						Node toNode = suc.getToNode();
						if (block.getNodes().contains(toNode))
							return toNode;
					}
				}				
			}
		}
		
		return null;
	}

	/**
	 * Find all nodes that are head of an edge with label=action
	 * and which has tail=from.
	 * @param side The side (left/right)
	 * @param from The from-node
	 * @param action The action name
	 * @return A set of choices
	 */
	public ArrayList<Node> getChoices(int side, Node from, String action) {
		ArrayList<Node> choices = new ArrayList<Node>();

		if (side == GameModel.LEFT_SIDE) {
			for (Successor suc : leftSuccessorMap.get(from)) {
				if (suc.getEdgeLabel().equals(action)) {
					choices.add(suc.getToNode());
				}
			}
		}
		else {
			for (Successor suc : rightSuccessorMap.get(from)) {
				if (suc.getEdgeLabel().equals(action)) {
					choices.add(suc.getToNode());
				}
			}			
		}

		return choices;
	}
	
	/**
	 * Returns the initial node of the left LTS
	 * @return The initial node of the left LTS
	 */
	public Node getInitialNodeLeft() {
		return initialNodeLeft;
	}

	/**
	 * Returns the initial node of the right LTS
	 * @return The initial node of the right LTS
	 */
	public Node getInitialNodeRight() {
		return initialNodeRight;
	}
	
}
