package dk.brics.bisim.model;

import java.util.ArrayList;

import dk.brics.bisim.equiv.CGNode;
import dk.brics.bisim.equiv.ConfigurationGraph;
import dk.brics.bisim.equiv.StateSpace;

/**
 * A local game model holding two GrappaModels. It also holds the state space and configuration
 * graph of the combined states. This is used by AIs but also to detect infinite games, in which 
 * case the defender is declared winner. In practice, we do not wish to wait forever, so we
 * define an infinite game as returning to an already seen configuration (s,t) of nodes from the
 * left and right LTS respectively. This can be justified by the fact that a perfect attacker
 * can always avoid returning to a previously visited state.
 * 
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class GameModel {

	public static final int FINITE_GAME = -1;
	
	public static final int NOBODYS_TURN    = 1;
	public static final int TURN_ATTACKER_1 = 2;
	public static final int TURN_ATTACKER_2 = 3;	
	public static final int TURN_DEFENDER   = 4;	
	
	public static final int LEFT_SIDE  = 5;
	public static final int RIGHT_SIDE = 6;
	
	private int turn;
	private int attackerSide;
	private int roundNumber;
	private String attackerAction;
	private LTSModel leftLTSModel, rightLTSModel;
	private boolean gameOver;
	private int infiniteGameOver;
	
	private StateSpace stateSpace;
	private boolean hasStateSpacePartition;
	private ConfigurationGraph configurationGraph;
	private ArrayList<CGNode> seenConfigurations;
	
	public GameModel(String leftInitialNodeName, String rightInitialNodeName) {
		attackerSide = 0;
		leftLTSModel  = new LTSModel(leftInitialNodeName);
		rightLTSModel = new LTSModel(rightInitialNodeName);	
	}

	/**
	 * Create a state space, partition it, and create a
	 * corresponding configuration graph.
	 */
	public void buildStateSpace() {
		/*** Create a state space based on the bisimulation type ***/
		stateSpace = new StateSpace(leftLTSModel.getInitialNode(), 
									rightLTSModel.getInitialNode(), 
									leftLTSModel.getSuccessorMap(),
									rightLTSModel.getSuccessorMap());
		hasStateSpacePartition = stateSpace.partition();
		configurationGraph = new ConfigurationGraph(stateSpace);
		resetSeenConfigurations();
	}
	
	/**
	 * @return The attackers last chosen action
	 */
	public String getAttackerAction() {		
		return attackerAction;
	}

	/**
	 * Sets the attackers action
	 * @param attackerAction An action
	 */
	public void setAttackerAction(String attackerAction) {
		this.attackerAction = attackerAction;
	}

	/**
	 * @return The turn (attacker/defender)
	 */
	public int getTurn() {
		return turn;
	}

	/**
	 * Sets the turn (attacker/defender)
	 * @param turn A turn
	 */
	public void setTurn(int turn) {
		this.turn = turn;
	}

	/**
	 * @return The current attacker side
	 */
	public int getAttackerSide() {
		return attackerSide;
	}

	/**
	 * Sets the attacker side.
	 * @param attackerSide A side
	 */
	public void setAttackerSide(int attackerSide) {
		this.attackerSide = attackerSide;
	}

	/**
	 * @return The grappa model of the left LTS
	 */
	public LTSModel getLeftLTSModel() {
		return leftLTSModel;
	}

	/**
	 * @return The grappa model of the right LTS
	 */
	public LTSModel getRightLTSModel() {
		return rightLTSModel;
	}

	/**
	 * Reset LTS models.
	 */
	public void reset() {
		leftLTSModel.reset();
		rightLTSModel.reset();
		resetSeenConfigurations();
	}
	
	/**
	 * Reset the set of seen configurations and add
	 * the initial configuration.
	 */
	private void resetSeenConfigurations() {
		roundNumber = 0;
		infiniteGameOver = FINITE_GAME;
		/*** Add initial configuration ***/
		seenConfigurations = new ArrayList<CGNode>();
		CGNode cur = configurationGraph.getNode(leftLTSModel.getInitialNode(),
												rightLTSModel.getInitialNode());
		seenConfigurations.add(cur);		
	}

	/**
	 * @return Game over status
	 */
	public boolean isGameOver() {
		return gameOver;
	}
	
	/**
	 * @return If an infinite game has been won by the defender
	 */
	public int getInfiniteGameOver() {
		return infiniteGameOver;
	}	

	/**
	 * Calculates if game is over.
	 */
	public void setGameOver() {
		if (turn == TURN_ATTACKER_2) {
			/*** Get defender's LTS ***/
			if (attackerSide == LEFT_SIDE) {
				gameOver = !rightLTSModel.isActionEnabled(attackerAction);
			}
			else {
				gameOver = !leftLTSModel.isActionEnabled(attackerAction);			
			}
		}
		else if (turn == TURN_DEFENDER) {
			/*** Add current configuration ***/
			CGNode cur = configurationGraph.getNode(leftLTSModel.getCurrentNode(),
													rightLTSModel.getCurrentNode());
			if (seenConfigurations.contains(cur)) {
				/** Current configuration was already seen **/
				gameOver = true;
				infiniteGameOver = seenConfigurations.indexOf(cur);
			}
			else {
				/** Current configuration wasn't seen already **/
				seenConfigurations.add(cur); 				
				gameOver = (leftLTSModel.noActionEnabled() &&
				   	    	rightLTSModel.noActionEnabled());				
			}
		}
		else 
			gameOver = false;		
	}

	public int incAndGetRoundNumber() {
		return roundNumber++;
	}

	public StateSpace getStateSpace() {
		return stateSpace;
	}

	public ConfigurationGraph getConfigurationGraph() {
		return configurationGraph;
	}

	public boolean hasStateSpacePartition() {
		return hasStateSpacePartition;
	}

}
