package dk.brics.bisim.model;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;

import dk.brics.bisim.equiv.StrongSuccessor;
import dk.brics.bisim.equiv.Successor;
import dk.brics.bisim.equiv.WeakSuccessor;

import att.grappa.Edge;
import att.grappa.Graph;
import att.grappa.GrappaSupport;
import att.grappa.Node;
import att.grappa.Parser;

/**
 * A local data model holding a Grappa graph. It also holds a map of successors of nodes. This map
 * can be weak or strong depending on the bisimulation type.
 * 
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class LTSModel {

	private Graph graph;
	private Node currentNode, toNode;
	private ArrayList<Edge> currentEdges;
	private String initialNodeName;
	private HashMap<Node,ArrayList<Successor>> successorMap;
	
	/**
	 * Constructs an LTS model given the name of the
	 * initial node in the graph.
	 * @param initialNodeName The name of the initial node
	 */
	public LTSModel(String initialNodeName) {
		this.initialNodeName = initialNodeName;
		currentEdges = new ArrayList<Edge>();
	}	
	
	/**
	 * Used by the CCS Simulator to always create a strong successor map.
	 * @param dot
	 * @param dotPath
	 * @return true if the graph has node elements
	 */
	public boolean makeGraphFromDot(String dot, String dotPath) {
		return makeGraphFromDot(dot, dotPath, OptionsModel.BISIMULATION_STRONG);
	}
	
	/**
	 * First construct a dot graph, then pass it to Grappa and try
	 * to make a Graph.
	 * @param dot String representation of a dot-file
	 * @param dotPath Path to dot program
	 * @param bisimulation Bisimulation type, used for creating a successor map
	 * @return true if the graph has node elements
	 */
	public boolean makeGraphFromDot(String dot, String dotPath, int bisimulation) {
		Reader reader = new StringReader(dot);

		Parser program = new Parser (reader, new PrintWriter(System.err));
		/***** Parse chosen file *****/
		try {
		    program.parse();
		} catch(Exception ex) {
		    System.err.println("Could not parse dot format.\n" + ex.getMessage());
		    System.exit(1);
		}
		
		/***** Make graph *****/
		graph = program.getGraph();
		graph.setErrorWriter(new PrintWriter(System.err,true));
		layoutGraph(dotPath);
		
		currentNode = graph.findNodeByName(initialNodeName);
		if (currentNode == null) {
			System.err.println("Could not find initial node!");
		}
		
		/*** Build successor map ***/
		if (bisimulation == OptionsModel.BISIMULATION_STRONG)
			successorMap = getStrongSuccessorMap();
		else
			successorMap = getWeakSuccessorMap();

		return graph.nodeElements().hasMoreElements();	
	}

	/**
	 * Attempts to layout a graph using the Graphviz dot program.
	 * @param dotPath Path to the dot executable
	 */
	private void layoutGraph (String dotPath) {
		System.out.print("Calling 'dot' to layout graph...");
		Process connector = null;
		List<String> command = new ArrayList<String>();
		command.add(dotPath);
		ProcessBuilder builder = new ProcessBuilder(command);
		try {
			connector = builder.start();
		} catch (IOException e) {
			System.out.println("Unable to start dot process. Program not found on the path:");
			System.out.println(dotPath);
		}
		
		if (GrappaSupport.filterGraph (graph, connector)) {
			System.out.println(" done!");
		}
		else {
			System.out.println("Failed to layout graph");
		}
	}

	/**
	 * @return The Grappa graph
	 */
	public Graph getGraph() {
		return graph;
	}	

	/**
	 * @return The current node
	 */
	public Node getCurrentNode() {
		return currentNode;
	}

	/**
	 * Called when an edge has been clicked by  
	 * the attacker and returns action name.
	 * @param edge Edge that has been mouse-clicked
     * @return The action name of the edge.
	 */
	public String getEdgeAction(Edge edge) {
		String action = null;
		
		if (edge.getTail().equals(currentNode)) {
			action = (String)edge.getAttributeValue("label");
		}

		return action;
	}
	
	/**
	 * Called when an edge has been clicked by  
	 * the defender.
	 * The edge action must match, even in the
	 * case of weak bisimulation.
	 * @param edge Edge that has been mouse-clicked
     * @param otherAction Other action that must be matched
     * @return True if the clicked edge has same name as otherAction
	 */
	public boolean getMatchingEdgeAction(Edge edge, String otherAction) {
		if (edge.getTail().equals(currentNode) &&
				((String)edge.getAttributeValue("label")).equals(otherAction)) {
			return true;
		}

		return false;
	}
	
	/**
	 * Called when a node has been clicked by the
	 * attacker.
	 * If the node has a unique edge from the current node,
	 * the action name of that edge is returned.
	 * The successor map cannot be used here, since in the case of a
	 * weak bisimulation, a node will always have a tau self-loop,
	 * which is only eligible for a defender, however.
	 * @param node The node clicked.
	 * @return The name of the action of the unique edge between current
	 * node and clicked, or null if it does not exist.
	 */
	public String getUniqueNodeAction(Node node) {
		String action = null;
		boolean found = false;
		
		Enumeration inEdges = node.inEdgeElements();	
		while (inEdges.hasMoreElements()) {
			Edge edge = (Edge)inEdges.nextElement();
			Node fromNode = edge.getTail();
			if (fromNode.equals(currentNode)) {
				if (found) {
					return null; // No unique action can be chosen
				}
				else {
					action = (String)edge.getAttributeValue("label");			
					found = true;
				}
			}			
		}
				
		return action;
	}

	/**
	 * Called when a node has been clicked by the
	 * defender.
	 * If the node has an in-edge from the current node
	 * that matches attackersAction, return true.
	 * In case of weak bisimulation, there must exist a
	 * legal path from the current node to the clicked
	 * node (recorded in the successor map).
	 * A legal path is (tau)*(attackersAction)(tau)*
	 * @param node The node clicked
	 * @param attackersAction Other action that must be matched
	 * @return Whether there exists an edge between current
	 * node and clicked with name otherAction
	 */
	public boolean getMatchingNodeAction(Node node, String attackersAction) {
		for (Successor suc : successorMap.get(currentNode)) 
			if (suc.getToNode().equals(node) &&
				suc.getEdgeLabel().equals(attackersAction))
				return true;
		
		return false;
	}

	private void step() {
		currentNode.setAttribute("color", "white");
		currentNode = toNode;
		currentNode.setAttribute("color", "red");
		for (Edge currentEdge : currentEdges) {
			currentEdge.setAttribute("color", "blue");
			currentEdge.setAttribute("fontcolor", "blue");
		}		
		graph.repaint();
	}

	/**
	 * Reset the displayed automaton to its initial state.
	 */
	public void reset() {
		resetCurrentEdges();
		stepTo(initialNodeName);
	}

	/**
	 * Step to a given node in the graph.
	 * @param nodeName The name of the node to step to
	 */
	public void stepTo(String nodeName) {
		toNode = graph.findNodeByName(nodeName);
		step ();
	}
	
	/**
	 * Step to a given node, and color
	 * the edge used.
	 * @param nodeName The name of the node to step to
	 * @param edgeLabel The label of the edge to color
	 */
	public void stepTo(String nodeName, String edgeLabel) {
		toNode = graph.findNodeByName(nodeName);
		currentEdges = findCurrentEdges(edgeLabel);
		step();
	}
	
	/**
	 * Finds a given edge path from current node to to-node.
	 * Assumes these nodes have been set up.
	 * @param edgeLabel The edge label
	 * @return The edge, or null if it wasn't found
	 */
	private ArrayList<Edge> findCurrentEdges(String edgeLabel) {
		ArrayList<Edge> result = new ArrayList<Edge>();
		ArrayList<Successor> cMap = successorMap.get(currentNode);
		
		for (Successor suc : cMap) {
			if (suc.getToNode().equals(toNode) &&
				suc.getEdgeLabel().equals(edgeLabel)) {
				for (String name : suc.getEdgeNames()) {
					result.add(graph.findEdgeByName(name));
				}
			}
		}
		
		return result;
	}
	
	/**
	 * Reset current edges to default color (black).
	 */
	public void resetCurrentEdges() {
		for (Edge edge : currentEdges) {
			edge.setAttribute("color", "black");
			edge.setAttribute("fontcolor", "black");
		}
		currentEdges.clear();
		graph.repaint();
	}
	
	/**
	 * Check to see if there exists an outgoing edge from
	 * current node that has label = action.
	 * @param action Name of edge label to match
	 * @return If a matching edge exists
	 */
	public boolean isActionEnabled(String action) {
		for (Successor suc : successorMap.get(currentNode)) 
			if (suc.getEdgeLabel().equals(action))
				return true;
		
		return false;
	}
	
	/**
	 * @return If current node has no (strong) outgoing edges
	 */
	public boolean noActionEnabled() {
		return getStrongChoices().isEmpty();
	}

	/**
	 * Return the number of outgoing edges from 
	 * the current node.
	 * @return The number of outgoing edges from the
	 * current node
	 */
	public int getNumberOfStrongChoices() {
		return getStrongChoices().size();
	}
	
	/**
	 * Used by an attacker to get all choices from 
	 * the current node that are not weak.
	 * @return The outgoing edges from the current node
	 */
	public ArrayList<Successor> getStrongChoices() {
		ArrayList<Successor> choices = new ArrayList<Successor>();
		
		for (Successor suc : successorMap.get(currentNode)) {
			if (!suc.isWeak())
				choices.add(suc);
		}
		
		return choices;
	}
	
	/**
	 * @return The initial node of the Grappa graph
	 */
	public Node getInitialNode() {
		return graph.findNodeByName(initialNodeName);
	}
	
	/**
	 * Makes a map from each node to its direct successors.
	 * @return A strong successor map
	 */
	private HashMap<Node,ArrayList<Successor>> getStrongSuccessorMap() {
		HashMap<Node,ArrayList<Successor>> successorMap = new HashMap<Node,ArrayList<Successor>>();
		ArrayList<Successor> toMap;
	
		Enumeration ne = graph.nodeElements();	
		while (ne.hasMoreElements()) {
			Node node = (Node)ne.nextElement();
			toMap = new ArrayList<Successor>();

			/*** Fill up the to-map ***/
			Enumeration outEdges = node.outEdgeElements();
			while (outEdges.hasMoreElements()) {
				Edge edge = (Edge)outEdges.nextElement();
				Node head = edge.getHead();
				String action = (String)edge.getAttributeValue("label");

				toMap.add(new StrongSuccessor(action, edge.getName(), head));
			}
			
			/*** Attach it to this node ***/
			successorMap.put(node,toMap);
		}

		return successorMap;
	}
	       
	/**
	 * Makes a map from each node to its weak successors.
	 * A variation of Warshall's transitive closure algorithm:
	 * Starts with the strong succesor map (direct successors),
	 * add tau self-loops for each node, 
	 * and then for each node-triple (i,j,k):
	 * for each pair of edges (i--e1-->k) and (k--e2-->j):
	 * if e1=tau, add an edge (i--e2-->j)
	 * else if e2=tau, add an edge (i--e1-->j). 
	 * @return A weak successor map
	 */
	private HashMap<Node,ArrayList<Successor>> getWeakSuccessorMap() {
		/*** Start with the strong map ***/
		HashMap<Node,ArrayList<Successor>> successorMap = getStrongSuccessorMap();

		/*** Add tau self-loops ***/
		Enumeration ne = graph.nodeElements();
		while (ne.hasMoreElements()) {
			Node n = (Node)ne.nextElement();
			ArrayList<Successor> nMap = successorMap.get(n);
			/*** Create a weak successor ***/
			WeakSuccessor selfLoop = new WeakSuccessor("tau", n);
			if (!nMap.contains(selfLoop)) {
				nMap.add(selfLoop);
				successorMap.put(n, nMap);
			}
		}
		
		/*** Unstable means an edge has been added and we reconsider
		 *   all node-triples ***/
		boolean unstable = true;
		while (unstable) {
			unstable = false;
			Enumeration ie = graph.nodeElements();	
			while (ie.hasMoreElements()) {
				Node i = (Node)ie.nextElement();			
				Enumeration je = graph.nodeElements();
				while (je.hasMoreElements()) {
					Node j = (Node)je.nextElement();
					Enumeration ke = graph.nodeElements();
					while (ke.hasMoreElements()) {
						Node k = (Node)ke.nextElement();
						ArrayList<Successor> iMap = successorMap.get(i);
						ArrayList<Successor> kMap = successorMap.get(k);
					
						/*** Get all successors from i to k ***/
						ArrayList<Successor> iToK = filterSuccessors(iMap, k);
						/*** Get all successors from k to j ***/
						ArrayList<Successor> kToJ = filterSuccessors(kMap, j);
						for (Successor s1 : iToK) {
							for (Successor s2 : kToJ) {
								String e1 = s1.getEdgeLabel();
								String e2 = s2.getEdgeLabel();
								
								/*** If one edge is tau and head and
								 *   tail of that edge is not the same
								 *   node, add the other edge.
								 *   If neither edge is tau, don't add a new weak edge.
								 */
								WeakSuccessor suc;
								if (e1.equals("tau")) {
									suc = new WeakSuccessor(e2, j);
									if (!iMap.contains(suc)) {
										/**
										 * Concatenate edge names from the two other 
										 * successors.
										 */
										suc.addEdgeNames(s1.getEdgeNames());
										suc.addEdgeNames(s2.getEdgeNames());
										iMap.add(suc);
										unstable = true;
									}
								}
								else if (e2.equals("tau")) {
									suc = new WeakSuccessor(e1, j);
									if (!iMap.contains(suc)) {
										/**
										 * Concatenate edge names from the two other 
										 * successors.
										 */
										suc.addEdgeNames(s1.getEdgeNames());
										suc.addEdgeNames(s2.getEdgeNames());
										iMap.add(suc);
										unstable = true;
									}
								}
							}
						}
						/*** Update with the new value ***/
						successorMap.put(i, iMap);
					}
				}
			}
		}

		return successorMap;		
	}	
	
	/**
	 * Get the labels of in-edges of node 'to'. 
	 * Used for computing a weak successor map.
	 * @param toMap
	 * @param to
	 * @return The labels of in-edges of node 'to'
	 */
	private ArrayList<Successor> filterSuccessors(ArrayList<Successor> toMap, Node to) {
		ArrayList<Successor> edgeLabels = new ArrayList<Successor>();
		
		for (Successor suc : toMap) {
			if (suc.getToNode().equals(to))
				edgeLabels.add(suc);
		}
		
		return edgeLabels;
	}

	/**
	 * @return The successor map
	 */
	public HashMap<Node,ArrayList<Successor>> getSuccessorMap() {
		return successorMap;
	}

	public ArrayList<Successor> getCurrentSuccessors() {
		return successorMap.get(currentNode);
	}
	
} 


