package dk.brics.bisim.model;

/**
 * Data model for the game options.
 * Used to hold the game options chosen by the user.<p>
 * 
 * These are:
 * <ul>
 * <li>Number of players</li>
 * <li>Local or remote game</li>
 * <li>Client/server role</li>
 * <li>Possibly a host</li>
 * <li>A player role (attacker/defender)</li>
 * <li>The left and right LTSs</li>
 * <li>The bisimulation type (strong/weak)</li>
 * </ul>
 *
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class OptionsModel {

	private static final String CHOOSE_PLAYERS       = "Please choose number of players";
	private static final String CHOOSE_ROLE          = "Please choose a role";
	private static final String CHOOSE_HOST          = "Please choose a host";
	private static final String CHOOSE_CLIENT_SERVER = "Please choose server or client";
	private static final String CHOOSE_LEFT_LTS      = "Please choose left LTS";
	private static final String CHOOSE_RIGHT_LTS     = "Please choose right LTS";
	private static final String CHOOSE_BISIMULATION  = "Please choose bisimulation type";
	
	public static final int ONE_PLAYER        = 1;
	public static final int TWO_PLAYER_LOCAL  = 2;
	public static final int TWO_PLAYER_REMOTE = 3;
	
	public static final int ATTACKER = 4;
	public static final int DEFENDER = 5;
	public static final int BOTH     = 6;
	
	public static final int CLIENT = 7;
	public static final int SERVER = 8;
	public static final int LOCAL  = 9;
	
	public static final int BISIMULATION_STRONG = 10;
	public static final int BISIMULATION_WEAK   = 11;	
	
	private int players, 
	            role, 
	            clientServer, 
	            bisimulationType;
	private String leftLTSPath, 
				   rightLTSPath,
				   leftLTSFile, 
				   rightLTSFile,
				   host;
	
	/**
	 * Creates an options model.
	 */
	public OptionsModel() {
		players = 0;
		role = BOTH;
		clientServer = LOCAL;
		bisimulationType = 0;
		
		leftLTSPath = "";
		rightLTSPath = "";
		host = "";
	}

	/**
	 * Checks if the user has chosen game properties.
	 * @return Status string which is empty in case properties have been chosen
	 */
	public String hasProperties() {
		String mes = "";
		
		if (players == ONE_PLAYER) {
			/*** 1p must choose: role, LTSs, bisimulation type ***/
			if (role == BOTH) {
				mes = CHOOSE_ROLE;
			}
			else if (leftLTSPath.equals("")) {
				mes = CHOOSE_LEFT_LTS;
			}
			else if (rightLTSPath.equals("")) {
				mes = CHOOSE_RIGHT_LTS;	
			}
			else if (bisimulationType == 0) {
				mes = CHOOSE_BISIMULATION;
			}
		}
		else if (players == TWO_PLAYER_LOCAL) {			
			/*** 2p local must choose: LTSs, bisimulation type ***/
			if (leftLTSPath.equals("")) {
				mes = CHOOSE_LEFT_LTS;
			}
			else if (rightLTSPath.equals("")) {
				mes = CHOOSE_RIGHT_LTS;	
			}
			else if (bisimulationType == 0) {
				mes = CHOOSE_BISIMULATION;
			}		
		}
		else if (players == TWO_PLAYER_REMOTE) {
			if (clientServer == SERVER) {
				/*** Server must choose: role, LTSs, bisimulation type ***/
				if (role == BOTH) {
					mes = CHOOSE_ROLE;
				}
				else if (leftLTSPath.equals("")) {
					mes = CHOOSE_LEFT_LTS;
				}
				else if (rightLTSPath.equals("")) {
					mes = CHOOSE_RIGHT_LTS;	
				}
				else if (bisimulationType == 0) {
					mes = CHOOSE_BISIMULATION;
				}
			}
			else if (clientServer == CLIENT) {
				/*** Client must choose: Host ***/
				if (host.equals("")) {
					mes = CHOOSE_HOST;
				}
			}
			else mes = CHOOSE_CLIENT_SERVER; // Client/Server type not chosen
		}
		else mes = CHOOSE_PLAYERS; // Number of players not chosen		
	
		return mes;
	}

	/**
	 * @return The bisimulation type (strong/weak)
	 */
	public int getBisimulationType() {
		return bisimulationType;
	}

	/**
	 * Set the bisimulation type.
	 * @param bisimulationType Strong or weak as defined in constant field
	 * values
	 */
	public void setBisimulationType(int bisimulationType) {
		this.bisimulationType = bisimulationType;
	}

	/**
	 * @return The left LTS file
	 */
	public String getLeftLTSFile() {
		return leftLTSFile;
	}

	/**
	 * Set the left LTS file
	 * @param leftLTSFile A file name
	 */
	public void setLeftLTSFile(String leftLTSFile) {
		this.leftLTSFile = leftLTSFile;
	}

	/**
	 * @return The path of the left LTS
	 */
	public String getLeftLTSPath() {
		return leftLTSPath;
	}

	/**
	 * Set the path of the left LTS.
	 * @param leftLTSPath A path name
	 */
	public void setLeftLTSPath(String leftLTSPath) {
		this.leftLTSPath = leftLTSPath;
	}

	/**
	 * @return The right LTS file
	 */
	public String getRightLTSFile() {
		return rightLTSFile;
	}

	/**
	 * Set the right LTS file.
	 * @param rightLTSFile A file name
	 */
	public void setRightLTSFile(String rightLTSFile) {
		this.rightLTSFile = rightLTSFile;
	}

	/**
	 * @return The path name of the right LTS
	 */
	public String getRightLTSPath() {
		return rightLTSPath;
	}

	/**
	 * Set the path of the right LTS.
	 * @param rightLTSPath A path name
	 */
	public void setRightLTSPath(String rightLTSPath) {
		this.rightLTSPath = rightLTSPath;
	}

	/**
	 * @return The client/server type as defined in
	 * constant field values
	 */
	public int getClientServer() {
		return clientServer;
	}

	/**
	 * Set the client/server type.
	 * @param clientServer A constant field value
	 */
	public void setClientServer(int clientServer) {
		this.clientServer = clientServer;
	}

	/**
	 * @return The number of players
	 */
	public int getPlayers() {
		return players;
	}

	/**
	 * Set the number of players.
	 * @param players A constant field value
	 */
	public void setPlayers(int players) {
		this.players = players;
	}

	/**
	 * @return The role
	 */
	public int getRole() {
		return role;
	}

	/**
	 * Set the role.
	 * @param role A constant field value
	 */
	public void setRole(int role) {
		this.role = role;
	}

	/**
	 * @return The host
	 */
	public String getHost() {
		return host;
	}

	/**
	 * Set the host.
	 * @param host A host name (IP address)
	 */
	public void setHost(String host) {
		this.host = host;
	}
	
	/**
	 * This is a hack to ensure that a two player local game
	 * gets role=both.
	 */
	public void checkRole() {
		if (players == TWO_PLAYER_LOCAL)
			role = BOTH;
	}

}
