package dk.brics.bisim.model;

import java.util.ArrayList;
import java.util.HashMap;

import att.grappa.Edge;
import att.grappa.Graph;

import dk.brics.bisim.equiv.Successor;

/**
 * A data model for the CCS Simulator.
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class SimulatorModel {

	private LTSModel ltsModel;
	private Edge emphasizedEdge;
	private static final String EMP_COLOR = "deepskyblue";
		
	/**
	 * Input actions are mapped by (unique) name to label.
	 */
	private HashMap<String,String> inputActions;
	
	/**
	 * Output actions are mapped by (unique) name to label.
	 */
	private HashMap<String,String> outputActions;
	
	private HashMap<String,String> allActions;
	
	/**
	 * Tau actions are saved by name (label is always 'tau').
	 */
	private ArrayList<String> tauActions;
	
	private DotModel dotModel;

	public SimulatorModel(DotModel dotModel) {
		inputActions = new HashMap<String,String>();
		outputActions = new HashMap<String,String>();
		allActions = new HashMap<String,String>();
		tauActions = new ArrayList<String>();
		this.dotModel = dotModel;
	}

	public LTSModel getLTSModel() {
		return ltsModel;
	}
	
	private void setActions() {
		ArrayList<Successor> currentActions = ltsModel.getCurrentSuccessors();

		inputActions.clear();
		outputActions.clear();
		allActions.clear();
		tauActions.clear();
		
		for (Successor suc : currentActions) {
			String action = suc.getEdgeLabel();
			String name = suc.getEdgeNames().get(0);
			if (action.startsWith("'")) {
				outputActions.put(name, action);		
			}
			else if (action.equals("tau")) {
				tauActions.add(name);
			}
			else {
				inputActions.put(name, action);
			}
			allActions.put(name, action);
		}
	}
	
	public HashMap<String,String> getOutputActions() {
		return outputActions;
	}
	
	public HashMap<String,String> getInputActions() {
		return inputActions;
	}
	
	public boolean hasTau() {
		return !tauActions.isEmpty();
	}

	/**
	 * Performs a randomstep from the actions available
	 * in the map.
	 * @param actions A HashMap of actions
	 * @return The chosen action
	 */
	private String stepRandom(HashMap<String,String> actions) {
		String action = null;
		String name;
		
		name = (String) actions.keySet().toArray()[(int) (Math.random()*actions.size())];
		action = actions.get(name);
		stepTo(name, action);
		
		return action;	
	}
	
	/**
	 * Choose a random (available) action.
	 * @return The chosen action
	 */
	public String stepRandomAll() {
		return stepRandom(allActions);
	}
	
	/**
	 * Choose a random (available) input action.
	 * @return The chosen action
	 */
	public String stepRandomInput() {
		return stepRandom(inputActions);
	}
	
	/**
	 * Choose a random (available) output action.
	 * @return The chosen action
	 */
	public String stepRandomOutput() {
		return stepRandom(outputActions);
	}
	
	/**
	 * Choose a random (available) tau action.
	 */
	public void stepRandomTau() {
		String name = tauActions.get((int) (Math.random()*tauActions.size()));
		stepTo(name, "tau");	
	}
	
	public boolean createLTS(String fileName, String filePath, String dotPath) {
		String initialNodeName  = fileName.substring(0, fileName.indexOf('.'));
		ltsModel = new LTSModel(initialNodeName);

		return ltsModel.makeGraphFromDot(dotModel.getDotFromGraph(filePath), dotPath);
	}

	/**
	 * Reset the simulator model and its underlying 
	 * LTS model.
	 */
	public void reset() {
		ltsModel.reset();
		setActions();
	}

	/**
	 * Step in the LTS model.
	 * @param edgeName
	 * @param edgeLabel
	 */
	public void stepTo(String edgeName, String edgeLabel) {
		String nodeName = ltsModel.getGraph().findEdgeByName(edgeName).getHead().getName();
		ltsModel.resetCurrentEdges();
		ltsModel.stepTo(nodeName, edgeLabel);
		setActions();
	}
	
	/**
	 * Step in the LTS model.
	 * @param nodeName
	 * @param edgeLabel
	 */
	public void stepToNode(String nodeName, String edgeLabel) {
		ltsModel.resetCurrentEdges();
		ltsModel.stepTo(nodeName, edgeLabel);
		setActions();
	}

	/**
	 * Emphasize an edge when mouse hovers over an action name.
	 * @param edgeName Name of the edge
	 * @param edgeLabel Label of the edge
	 */
	public void emphasizeEdge(String edgeName, String edgeLabel) {
		for (Successor suc : ltsModel.getCurrentSuccessors()) {
			String sucEdgeName = suc.getEdgeNames().get(0);
			if (sucEdgeName.equals(edgeName) &&
				suc.getEdgeLabel().equals(edgeLabel)) {
				/**
				 * Just take the first edge that matches.
				 * There will only be more if there are more 
				 * edges with same head and tail nodes and labels.
				 */
				Graph g = ltsModel.getGraph();
				Edge e = g.findEdgeByName(sucEdgeName);
				if (e != null) {
					emphasizedEdge = e;
					emphasizedEdge.setAttribute("color", EMP_COLOR);
					emphasizedEdge.setAttribute("fontcolor", EMP_COLOR);
					g.repaint();
				}
			}
		}
	}
	
	/**
	 * De-emphasize an edge when mouse exits an action name.
	 * @param nodeName Name of the to-node
	 * @param edgeLabel Label of the edge
	 */
	public void deemphasizeEdge(String nodeName, String edgeLabel) {
		emphasizedEdge.setAttribute("color", "black");
		emphasizedEdge.setAttribute("fontcolor", "black");		
		ltsModel.getGraph().repaint();
	}
	
}
