package dk.brics.bisim.view;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Image;
import java.io.ByteArrayInputStream;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledDocument;

import dk.brics.bisim.JarResources;

import att.grappa.Graph;
import att.grappa.GrappaAdapter;

/**
 * The game view.<p>
 * 
 * The game view consists of two LTS panels, left and right, a status panel below, and
 * a chat and button panel above.
 * 
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class GameView extends JPanel {

	private static final long serialVersionUID = 1L;

	/**
	 * A local game does not have a chat function.
	 */
	public static final int GAME_VIEW_LOCAL  = 1;
	
	/**
	 * A server gets both a chat function and a reset button. 
	 */
	public static final int GAME_VIEW_SERVER = 2;
	
	/**
	 * A client does not have a reset button. 
	 */
	public static final int GAME_VIEW_CLIENT = 3;
	
	/**
	 * A panel for the left LTS.
	 */
	private LTSPanel leftLTSPanel;
	
	/**
	 * A panel for the right LTS. 
	 */
	private LTSPanel rightLTSPanel;
	
	/**
	 * The upper (north) panel of a game view.  
	 */
	private GameViewUpperPanel upperPanel;

	/**
	 * A scroll pane for the trace. 
	 */
	private JScrollPane traceScrollPane;

	/**
	 * A text pane inside the trace scroll pane.
	 */
	private JTextPane trace;
	
	public static final String STYLE_STATE    = "stateStyle";
	public static final String STYLE_ACTION   = "actionStyle";
	public static final String STYLE_COMMAND  = "commandStyle";
	public static final String STYLE_STATUS   = "statusStyle";
	public static final String STYLE_GAMEOVER = "gameoverStyle";
	
	private Image winnerImage, loserImage;
	private ImageIcon winnerImageIcon, loserImageIcon;
	
	private Dimension graphDimension, traceDimension;
	private int width, upperHeight;
	
	/**
	 * Constructs a game view but does not construct the lower panel.
	 * This is deferred until the player has chosen options.
	 * @param width Width of the main window
	 * @param height Height of the main window
	 * @throws BadLocationException Caused by using JTextPane. Should never be thrown
	 */
	public GameView(int width, int height) throws BadLocationException {
		this.width = width;
		int lowerHeight = 100;
		int graphHeight = height - (2 * lowerHeight);
		upperHeight = 60;
		
		graphDimension = new Dimension(this.width / 2, graphHeight);
		traceDimension = new Dimension(this.width, lowerHeight);
		
		/*** Create LTS panels ***/
		leftLTSPanel  = new LTSPanel(graphDimension);
		rightLTSPanel = new LTSPanel(graphDimension);
		    
	    /*** Create trace panel ***/
	    createImages();
		traceScrollPane = new JScrollPane();
	    traceScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
	    traceScrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	    resetTrace();	
	    
	    /*** Place components ***/	    
		setLayout(new BorderLayout());
		add(leftLTSPanel,    BorderLayout.WEST);
		add(rightLTSPanel,   BorderLayout.EAST);
		add(traceScrollPane, BorderLayout.SOUTH);
	}

	/**
	 * Load gif images from the jar file.
	 */
	private void createImages() {
		byte[] buffer = JarResources.getResource(JarResources.JAR_FILE, JarResources.WINNER_FILE);
	    try {
	        winnerImage = ImageIO.read(new ByteArrayInputStream(buffer));
	    } catch (IOException e) {
	    }
        if (winnerImage != null)
        	winnerImageIcon = new ImageIcon(winnerImage);
        
        buffer = JarResources.getResource(JarResources.JAR_FILE, JarResources.LOSER_FILE);
	    try {
	        loserImage = ImageIO.read(new ByteArrayInputStream(buffer));
	    } catch (IOException e) {
	    }
        if (loserImage != null)
        	loserImageIcon = new ImageIcon(loserImage);
	}	
	
	/**
	 * Creates the lower panel of a game view, based
	 * on game options
	 * @param gameViewType Type of game view
	 */
	public void createUpperPanel(int gameViewType) {
		if (upperPanel != null)
			remove(upperPanel);
		upperPanel = new GameViewUpperPanel(gameViewType, width, upperHeight);
		add(upperPanel, BorderLayout.NORTH);
	}
		
	/**
	 * Clear the text in the trace scroll pane.
	 * @throws BadLocationException Caused by using JTextPane. Should never be thrown
	 */
	public void resetTrace() throws BadLocationException {
		trace = new JTextPane();		
		insertTraceStyles();        
		trace.getDocument().insertString(0, "Trace:", null);
	    trace.setEditable(false);
	    traceScrollPane.setPreferredSize(traceDimension);
	    traceScrollPane.setViewportView(trace);
	}

	/**
	 * Generate the different styles used for
	 * writing in the trace text pane.
	 */
	private void insertTraceStyles() {
		Style style;
		
		/** State style is red and bold **/
        style = trace.getStyledDocument().addStyle(STYLE_STATE, null);
        StyleConstants.setForeground(style, Color.RED);
        StyleConstants.setBold(style, true);

        /** Action style is blue and bold **/
        style = trace.getStyledDocument().addStyle(STYLE_ACTION, null);
        StyleConstants.setForeground(style, Color.BLUE);
        StyleConstants.setBold(style, true);
		
        /** Command style is bold **/
        style = trace.getStyledDocument().addStyle(STYLE_COMMAND, null);
        StyleConstants.setBold(style, true);
        
        /** Status style is gray **/
        style = trace.getStyledDocument().addStyle(STYLE_STATUS, null);
        StyleConstants.setForeground(style, Color.GRAY);

		/** Gameover style is bold **/
        style = trace.getStyledDocument().addStyle(STYLE_GAMEOVER, null);
        StyleConstants.setBold(style, true);
	}
		
	/**
	 * Display the titles of the two LTSs.
	 * @param leftLTSTitle Title of the left LTS
	 * @param rightLTSTitle Title of the right LTS
	 */
	public void displayLTSTitles(String leftLTSTitle, String rightLTSTitle) {
		leftLTSPanel.setTitle(leftLTSTitle);
		rightLTSPanel.setTitle(rightLTSTitle);
	}

	/**
	 * Display the two LTSs.
	 * @param leftLTS Left LTS
	 * @param rightLTS Right LTS
	 * @param ga A Grappa adapter
	 */
	public void displayLTSs (Graph leftLTS, Graph rightLTS, GrappaAdapter ga) {
		leftLTSPanel.displayLTS(leftLTS, ga);
		rightLTSPanel.displayLTS(rightLTS, ga);
	}

	/**
	 * @return The left LTS panel
	 */
	public LTSPanel getLeftLTSPanel() {
		return leftLTSPanel;
	}

	/**
	 * @return The right LTS panel
	 */
	public LTSPanel getRightLTSPanel() {
		return rightLTSPanel;
	}
	
	/**
	 * Update the trace with a message. Possibly auto-scroll to the
	 * bottom.
	 * @param message A message
	 * @param doAutoScroll Whether or not to auto-scroll
	 * @throws BadLocationException Should never be thrown
	 */
	public void updateTrace(String message, boolean doAutoScroll) throws BadLocationException {
		StyledDocument doc = trace.getStyledDocument();		
		updateTrace(message, doAutoScroll, doc, null);
	}

	/**
	 * Update the trace with a styled message. Possibly auto-scroll to the
	 * bottom. 
	 * @param message A message
	 * @param doAutoScroll Whether or not to auto-scroll
	 * @param styleName Name of the style to write in
	 * @throws BadLocationException Should never be thrown
	 */
	public void updateTrace(String message, boolean doAutoScroll, String styleName) throws BadLocationException {
		StyledDocument doc = trace.getStyledDocument();		
		Style style = doc.getStyle(styleName);
		updateTrace(message, doAutoScroll, doc, style);
	}

	private void updateTrace(String message, boolean doAutoScroll, StyledDocument doc, Style style) throws BadLocationException {
		if (doAutoScroll) {
			doc.insertString(doc.getLength(), 
							 new StringBuilder("\n").append(message).toString(),
							 style);			
			trace.setCaretPosition(doc.getLength());
		}
		else {
			doc.insertString(doc.getLength(), message, style);
		}
	}	
	
	/**
	 * Update the trace with a message and an icon (winner or loser).
	 * @param message A message
	 * @param styleName The style to write in
	 * @param winner Whether or not to use the winner icon
	 * @throws BadLocationException Should never be thrown
	 */
	public void updateTrace(String message, String styleName, boolean winner) throws BadLocationException {
		StyledDocument doc = trace.getStyledDocument();
		Style style = doc.getStyle(styleName);
		updateTrace(message, true, doc, style);
		if (winner) {
			trace.insertIcon(winnerImageIcon);
		}
		else {
			trace.insertIcon(loserImageIcon);
		}
	}	

	/**
	 * @return The lower panel of the game view
	 */
	public GameViewUpperPanel getLowerPanel() {
		return upperPanel;
	}

}

