package dk.brics.bisim.view;

import java.awt.*;
import java.io.ByteArrayInputStream;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.text.BadLocationException;

import dk.brics.bisim.JarResources;

/**
 * The main view class, comprising the options view and the game view.
 * @see dk.brics.bisim.view.OptionsView
 * @see dk.brics.bisim.view.GameView
 * @author Martin Mosegaard <a href="mailto:martinm@daimi.au.dk">martinm@daimi.au.dk</a>
 *
 */
public class MainView extends JFrame {

	private static final long serialVersionUID = 1L;

	/**
	 * The tabbed pane holding an options view and a game view. 
	 */
	private JTabbedPane tabbedPane;
	
	/**
	 * The options view. 
	 */
	private OptionsView optionsView;
	
	/**
	 * The game view.
	 */
	private GameView gameView;

	/**
	 * The help page.
	 */
	private HelpView helpView;
	
	/**
	 * The about page.
	 */
	private AboutView aboutView;
	
	/**
	 * The CCS simulator view.
	 */
	private SimulatorView simulatorView;
	
	/**
	 * Main window title. 
	 */
	private static final String TITLE = "The Bisimulation Game Game";

	/**
	 * Options view tab title.
	 */
	private static final String optionsViewString = "Options";
	
	/**
	 * Game view tab title.
	 */
	private static final String gameViewString = "Game";
	
	/**
	 * Help page tab title.
	 */
	private static final String helpViewString = "Help";

	/**
	 * About page tab title.
	 */
	private static final String aboutViewString = "About";	
	
	private static final String simulatorViewString = "CCS Simulator";
	
	private JPanel contentPane;
	private static int width, height;
	private static Dimension screenSize;
	
	private Image itVestImage;
	protected static ImageIcon itVestImageIcon;
	
	/**
	 * Constructs a view and tries to set the look & feel to the
	 * system's look & feel.
	 * @throws BadLocationException Should never be thrown
	 */
	public MainView() throws BadLocationException {
		super(TITLE);
		
		/*** Try to set system Look&Feel ***/
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());			
		} catch (Exception e) {
			/*** Could not set system L&F ***/
		} 
		
		setDimensions();
	    create();		
		setLocation(0, 0);
		setDefaultCloseOperation(EXIT_ON_CLOSE);
	    setVisible(true);
	}
	
	private void setDimensions() {
		screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		width = screenSize.width;
		height = screenSize.height - 100;		
		setSize(width, height);
		setResizable(false);
	}

	/**
	 * @throws BadLocationException Should never be thrown
	 */
	private void create() throws BadLocationException {
		/** Create icon **/
		byte[] buffer = JarResources.getResource(JarResources.JAR_FILE, JarResources.IT_VEST_FILE);
	    try {
	        itVestImage = ImageIO.read(new ByteArrayInputStream(buffer));
	    } catch (IOException e) {
	    }
        if (itVestImage != null)
        	itVestImageIcon = new ImageIcon(itVestImage);

		optionsView = new OptionsView();
		gameView = new GameView(width-30, height-50);
		helpView = new HelpView(width-50, height-100);
		aboutView = new AboutView(width-50, height-100);
		simulatorView = new SimulatorView(width-50, height-150);
		
		tabbedPane = new JTabbedPane();
		tabbedPane.add(optionsViewString, optionsView);
		tabbedPane.add(gameViewString, gameView);
		tabbedPane.add(helpViewString, helpView);
		tabbedPane.add(aboutViewString, aboutView);
		tabbedPane.add(simulatorViewString, simulatorView);
		tabbedPane.setEnabledAt(tabbedPane.indexOfTab(gameViewString), false);
		
	    contentPane = new JPanel(new FlowLayout());
		contentPane.add(tabbedPane);

		setContentPane(contentPane);		
	}	
		
	/**
	 * @return The options view
	 */
	public OptionsView getOptionsView() {
		return optionsView;
	}

	/**
	 * Enable the game view tab.
	 */
	public void enableGameView() {
		int i = tabbedPane.indexOfTab(gameViewString);
		int j = tabbedPane.indexOfTab(optionsViewString);

		tabbedPane.setEnabledAt(i, true);		
		tabbedPane.setSelectedIndex(i);	
		tabbedPane.setEnabledAt(j, false);
	}

	/**
	 * Enable the options view tab.
	 */
	public void enableOptionsView() {
		int i = tabbedPane.indexOfTab(gameViewString);
		int j = tabbedPane.indexOfTab(optionsViewString);

		tabbedPane.setEnabledAt(j, true);		
		tabbedPane.setSelectedIndex(j);	
		tabbedPane.setEnabledAt(i, false);
	}

	/**
	 * Creates a lower panel in the game view
	 * and returns the game view.
	 * @param gameViewType A game view type
	 * @return A game view according to the type
	 */
	public GameView getGameView(int gameViewType) {
		/*** Add specific components to game view ***/
		gameView.createUpperPanel(gameViewType);
		
		return gameView;
	}

	public SimulatorView getSimulatorView() {
		return simulatorView;
	}

}
